#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 10 13:38:56 2025

@author: villa
"""

from pynter.vasp.schemes.core import InputSets

  
class ElectronicStructureSchemes(InputSets):
    """
    Class to generate and write input files for different calculation schemes in VASP
    """

    def dielectric_properties_complete(self,scheme_name='dielectric-properties',hybrid=False):
        """
        Set up calculations for electronic and ionic contribution in the dielectric constant.
        If hybrid is set to True the LCALCEPS method is used, else the LEPSILON is used.
        """
        jobs = []
        vaspjob = self.dielectric_properties_electronic()
        jobs.append(vaspjob)
        
        if hybrid:
            vaspjob = self.dielectric_properties_ionic_lcalceps()
        else:
            vaspjob = self.dielectric_properties_ionic_lepsilon()
        jobs.append(vaspjob)
        
        return jobs


    def dielectric_properties_electronic(self,scheme_name='eps-electronic'):
        """
        Set calculation for electronic contribution to the dielectric constant (and also dielectric function).
        Uses 'LOPTICS' method in VASP.
        """
        val = {}
        for p in self.potcar:
            val[p.element] = p.nelectrons
        vaspjob = self.get_vaspjob(add_to_job_name='eps-el',add_to_job_path=scheme_name)    
        vaspjob.incar['NEDOS'] = 2000
        vaspjob.incar['LOPTICS'] = '.TRUE.'              
        nelect = sum([ val[el]*self.structure.composition.as_dict()[el] for el in self.structure.composition.as_dict()]) #number of electrons
        nbands = int((nelect/2 + len(self.structure.sites)/2)*3) # 3*VASP default NBANDS (NELECT/2 +NIONS/2)
        vaspjob.incar['NBANDS'] = nbands
        
        return vaspjob
        

    def dielectric_properties_ionic_lcalceps(self,scheme_name='eps-ionic-lcalceps'):
        """
        Set calculation for ionic contribution to the dielectric constant.
        Uses 'LCALCEPS' method in VASP, combined with 'IBRION=6'. Useful for Hybrid calculations where 'LEPSILON' method does not work.
        """
        vaspjob = self.get_vaspjob(add_to_job_name='eps-ion-lcal',add_to_job_path=scheme_name)
        vaspjob.incar['NEDOS'] = 2000
        vaspjob.incar['LCALCEPS'] = '.TRUE.'              
        vaspjob.incar['IBRION'] = 6
        vaspjob.incar['NSW'] = 100
 
        return vaspjob


    def dielectric_properties_ionic_lepsilon(self,scheme_name='eps-ionic-lepsilon'):
        """
        Set calculation for ionic contribution to the dielectric constant.
        Uses 'LEPSILON' method in VASP, combined with 'IBRION=8'. This method does not work with HSE functionals.
        """
        vaspjob = self.get_vaspjob(add_to_job_name='eps-ion-leps',add_to_job_path=scheme_name)
        vaspjob.incar['NEDOS'] = 2000
        vaspjob.incar['LEPSILON'] = '.TRUE.'              
        vaspjob.incar['IBRION'] = 8
        vaspjob.incar['NSW'] = 100
 
        return vaspjob
    
        
    def pbe_electronic_structure(self,kmesh_dos=3, kpoints_bs=None,scheme_name='PBE-el-str'):
        """
        Generates calculation scheme for electronic structure calculations for PBE (in general non hybrid functionals) Steps: \n
            '1-PBE-relax': Relaxation of atomic positions and cell volume \n
            '2-PBE-DOS': DOS calculation with 2000 number of gridpoints on which the DOS is evaluated (NEDOS=2000) and increased k-mesh, ISMEAR is set to -5.
            '3-PBE-BS': Bandstructure calculation
        Parameters
        ----------
        kmesh_dos : (Int), optional
            Multiplier for coefficients of the k-mesh for the DOS with respect to first step SCF calculation. The default is 3.
        kpoints_bs : (Pymatgen Kpoints object) , optional
            Pymatgen Kpoints object for the path in BS calculation. The default is None. If None the default high symmetry path \n
            from Pymatgen class HighSymmKpath is obtained from the input Structure with 10 points between high symm k-points. '
            
        Returns
        -------
        List of VaspJob objects
        """
        stepnames = ['1-PBE-relax','2-PBE-DOS','3-PBE-BS']
        jobs = []
        
        sn = 1 
        vaspjob = self.pbe_ionic_vol_rel(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)  
        
        sn = 2 
        vaspjob = self.pbe_dos(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        vaspjob.incar['LCHARGE'] = True
        jobs.append(vaspjob)

        sn = 3
        vaspjob = self.pbe_bs(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        vaspjob.incar['LCHARGE'] = True
        jobs.append(vaspjob)
        
        return jobs


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun  5 16:24:40 2023

@author: villa
"""

import os
import yaml

from pynter import get_config_from_default_file, get_vasp_defaults_from_default_file, get_lammps_defaults_from_default_file

def setup_config(subparsers):
    parser_configure = subparsers.add_parser('configure',help='Set up pynter configuration files')
    
    parser_configure.add_argument('-f','--file',help='Configure file ("config.yml","vasp.yml","lammps.yml")',
                               default=None,metavar='',dest='file_to_configure')
    parser_configure.add_argument('-i','--info',action='store_true',help='Print configuration files paths',
                               default=False,dest='info')
    parser_configure.set_defaults(func=create_config)
    
    return
  

def create_config(args):
    run_config(file_to_configure=args.file_to_configure,info=args.info)
    return
  
def run_config(file_to_configure=None,info=False):
    
    filename_config = 'config_v2.yml'
    filename_vasp = 'vasp.yml'
    filename_lammps = 'lammps.yml'
    filepath = os.path.join(os.getenv("HOME"),'.pynter')
    
    # print configuration files path
    if info:
        print('Path of general configuration file: %s' %(os.path.join(filepath,filename_config)))
        print('Path of vasp configuration file: %s' %(os.path.join(filepath,filename_vasp)))
        print('Path of lammps configuration file: %s' %(os.path.join(filepath,filename_lammps)))

        return

    # Set config.yml
    if file_to_configure in [None, filename_config]:
        print('\nSet up configuration file for pynter...\n')
        print('Path of configuration file is: %s' %filepath)
        if not os.path.exists(filepath):
            os.makedirs(filepath)
        
        filename = filename_config
        
        print('Begin setup\n')
        print('HPC setup:')
        hostname = input('Hostname: ')
        localdir = input('Local calculation directory: ')
        localdir = os.path.abspath(localdir) if localdir else None
        remotedir = input('Remote calculation directory: ')
        remotedir = os.path.normpath(remotedir) if remotedir else None
        
        print('\nDefault settings for job script setup: ')
        project_id = input('Project ID: ')
        project_id = project_id if project_id else None
        
        job_script_filename = input('Job script filename (default job.sh): ')
        
        job_script_filename = job_script_filename if job_script_filename else 'job.sh'
        
        config = get_config_from_default_file()
        config.update({'HPC': 
                          {'hostname': hostname,
                          'localdir': localdir,
                          'remotedir': remotedir}})
        
        config['job_script_filename'] = job_script_filename
        config['sbatch']['account'] = project_id
                
        with open(os.path.join(filepath,filename),'w+') as f:
            yaml.dump(config,f) 
            
        print(f'Configuration file {filename} written in {filepath}\n')    
        
        
    # set vasp.yml
    if file_to_configure in [None, filename_vasp]:
    
        filename = filename_vasp
        vasp_default = get_vasp_defaults_from_default_file()
        
        path_vasp_exe = input('Default VASP executable: ')
        path_vasp_exe = path_vasp_exe or None
        
        vasp_default['job_settings_default']['vasp_exe'] = path_vasp_exe
        
        with open(os.path.join(filepath,filename),'w+') as f:
            yaml.dump(vasp_default,f) 
        
        print(f'Vasp default file {filename} written in {filepath}')  

    # set lammps.yml
    if file_to_configure in [None, filename_lammps]:
    
        filename = filename_lammps
        lammps_default = get_lammps_defaults_from_default_file()
        
        input_filename = input('Default LAMMPS input filename (default:input.in):')
        input_filename = input_filename or 'input.in'

        data_filename = input('Default LAMMPS data filename (default:structure.data):')
        data_filename = data_filename or 'structure.data'

        log_filename = input('Default LAMMPS log filename (default:log.lammps):')
        log_filename = log_filename or 'log.lammps'
        
        lammps_default['input_filename'] = input_filename
        lammps_default['data_filename'] = data_filename
        lammps_default['log_filename'] = log_filename


        path_lammps_exe = input('Default LAMMPS executable: ')
        path_lammps_exe = path_lammps_exe or None
        
        lammps_default['job_settings_default']['lammps_exe'] = path_lammps_exe
        
        with open(os.path.join(filepath,filename),'w+') as f:
            yaml.dump(lammps_default,f) 
        
        print(f'Lammps default file {filename} written in {filepath}')  














#!/usr/bin/env python

#from pymatgen.ext.matproj import MPRester
from mp_api.client import MPRester


class  MPDatabase:
    
    def __init__(self,mp_id=None,API_KEY=None):
        """
        Class to retrieve data from Materials Project database

        Parameters
        ----------
        mp_id : (str), optional
            Materials-ID. The default is None.
        """
        
        self.mp_id = mp_id if mp_id else None
        self.API_KEY = API_KEY
    
    @property
    def mp_rester(self):
        return MPRester(self.API_KEY)
        
        
    def get_entries(self,
                    chemsys_formula_mpids,
                    compatible_only=True,
                    property_data=None,
                    conventional_unit_cell=False):
        """
        Get a list of ComputedEntries or ComputedStructureEntries corresponding
        to a chemical system, formula, or materials_id or full criteria.

        Parameters
        ----------
            chemsys_formula_id_criteria (str/dict): A chemical system
                (e.g., Li-Fe-O), or formula (e.g., Fe2O3) or materials_id
                (e.g., mp-1234) or full Mongo-style dict criteria.
            compatible_only (bool): Whether to return only "compatible"
                entries. Compatible entries are entries that have been
                processed using the MaterialsProject2020Compatibility class,
                which performs adjustments to allow mixing of GGA and GGA+U
                calculations for more accurate phase diagrams and reaction
                energies.
            property_data (list): Specify additional properties to include in
                entry.data. If None, no data. Should be a subset of
                supported_properties.
            conventional_unit_cell (bool): Whether to get the standard
                conventional unit cell.

        Returns:
            List of ComputedEntry or ComputedStructureEntry objects.
        """
        with MPRester(self.API_KEY) as mpr:
            entries = mpr.get_entries(
                                chemsys_formula_mpids=chemsys_formula_mpids,
                                compatible_only=compatible_only,
                                property_data=property_data,
                                conventional_unit_cell=conventional_unit_cell)
        return entries
    
    
    def get_entries_from_compositions(self,
                                    compositions,
                                    compatible_only=True,
                                    property_data=None,
                                    conventional_unit_cell=False):
        """
        Get a dictionary with compositions (strings) as keys and a list of ComputedEntries 
        or ComputedStructureEntries as values.

        Parameters
        ----------
            compositions (list): List of strings with compositions.
            stable_only (bool): Get phase with lowest E above hull.
            compatible_only (bool): 
                Whether to return only "compatible"
                entries. Compatible entries are entries that have been
                processed using the MaterialsProject2020Compatibility class,
                which performs adjustments to allow mixing of GGA and GGA+U
                calculations for more accurate phase diagrams and reaction
                energies.
            property_data (list):
                Specify additional properties to include in
                entry.data. If None, no data. Should be a subset of
                supported_properties.
            conventional_unit_cell (bool):
                Whether to get the standard conventional unit cell

        Returns:
            List of ComputedEntry or ComputedStructureEntry objects.
        """
        entries_dict = {}
        for comp in compositions:
            entries = self.get_entries(
                                    chemsys_formula_mpids=comp,
                                    compatible_only=compatible_only,
                                    property_data=property_data,
                                    conventional_unit_cell=conventional_unit_cell)
            entries_dict[comp] = entries
        
        return entries_dict
    

    def get_structure(self,final=True,conventional_unit_cell=False):
        """
        Get a Structure corresponding to a material_id.

        Parameters
        ----------
            material_id (str): Materials Project material_id (a string,
                e.g., mp-1234).
            final (bool): Whether to get the final structure, or the initial
                (pre-relaxation) structure. Defaults to True.
            conventional_unit_cell (bool): Whether to get the standard
                conventional unit cell

        Returns:
            Structure object.
        """
        with MPRester(self.API_KEY) as mpr:
            structure = mpr.get_structure_by_material_id(self.mp_id,final=final,conventional_unit_cell=conventional_unit_cell)
        return structure

    
    
    
    
    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun  5 16:33:14 2023

@author: villa
"""

import os
import time
from datetime import datetime
import schedule
from shutil import which

from monty.dev import requires

from pynter import SETTINGS, run_command
from pynter.hpc.ssh import get_rsync_to_hpc_command, get_rsync_from_hpc_command
from pynter.hpc.slurm import JobSettings
from pynter.jobs.core import get_job_from_directory

def setup_hpc(subparsers):
    
    parser_hpc = subparsers.add_parser('HPC',help='Interface with High Performance Computer')
    subparsers_hpc = parser_hpc.add_subparsers()

    parser_sync = subparsers_hpc.add_parser('sync',help='Sync data with HPC with rsync')
    setup_sync(parser_sync)
    
    parser_job = subparsers_hpc.add_parser('job',help='Sync and run job on HPC')
    setup_job(parser_job)
    
    return


def setup_sync(parser):
    hostname = SETTINGS['HPC']['hostname']
    
    parser.add_argument('-dry','--dry-run',action='store_true',help='Dry run with rsync (default: %(default)s)',default=False,dest='dry_run')
    parser.add_argument('-exc','--exclude',action='append',help='Files to exclude from sync (default: %(default)s)',
                            default=['core.*','WAVECAR'],metavar='',type=str,dest='exclude')
    parser.add_argument('-host','--hostname',help='ssh alias of the HPC (default: %(default)s)',default=hostname,metavar='',type=str,dest='hostname')
    parser.add_argument('-ldir','--localdir',help='Local directory (default: %(default)s)',default=os.getcwd(),metavar='',type=str,dest='localdir')
    parser.add_argument('-wdir','--remotedir',help='Remote directory (default: %(default)s)',default=None,metavar='',type=str,dest='remotedir')
    parser.add_argument('-per','--periodic',action='store_true',help='Sync every hour with HPC (default: %(default)s)',default=False,dest='periodic')
    parser.add_argument('-st','--sync-to-hpc',help='Sync to hpc',required=False,default=False,action='store_true',dest='sync_to_hpc')
    parser.add_argument('-sf','--sync-from-hpc',help='Sync from hpc',required=False,default=False,action='store_true',dest='sync_from_hpc')
    
    parser.set_defaults(func=run_sync)
    return



@requires(which("rsync"),
      "rsync needs to be installed, you can install it with 'sudo apt-get install rsync'.")
def run_sync(args):
    command = get_command(args)
    execute_command(command,periodic=args.periodic)
    return

def get_command(args):
    hostname = args.hostname
    remotedir = args.remotedir
    localdir = args.localdir
    dry_run = args.dry_run
    exclude = args.exclude
    if args.sync_from_hpc and args.sync_to_hpc:
        raise ValueError('Cannot sync from HPC and to HPC at the same time')
    if args.sync_from_hpc:
        cmd = get_rsync_from_hpc_command(
                                    hostname=hostname,
                                    remotedir=remotedir,
                                    localdir=localdir,
                                    exclude=exclude,
                                    dry_run=dry_run
                                    )
    elif args.sync_to_hpc:
        cmd = get_rsync_to_hpc_command(
                                    hostname=hostname,
                                    remotedir=remotedir,
                                    localdir=localdir,
                                    exclude=exclude,
                                    dry_run=dry_run
                                    )         
    else:
        raise ValueError('Either --sync-from-hpc or --sync-from-hpc need to be specified')
    
    return cmd

    
def execute_command(command,periodic=False):
    if periodic:
        homedir = os.getenv("HOME")
        wdir = os.path.join(homedir,'.pynter','.sync_hpc')
        if not os.path.exists(wdir):
            os.makedirs(wdir)
        os.chdir(wdir)
        command = command.split()
        
        def job(command):        
            stdout, stderr = run_command(command,printout=True)
            dtn = datetime.now()
            date = dtn.date().isoformat()
            hour = dtn.ctime().split()[3]
            path = os.path.join(wdir,date)
            if not os.path.exists(path):
            		os.mkdir(path)
            	
            outfile = os.path.join(path,'sync_'+ hour +'.out')
            errfile = os.path.join(path,'sync_'+ hour +'.err')
            
            with open(outfile,'w') as outf, open(errfile,'w') as errf:
                outf.write(stdout)
                errf.write(stderr)
        
        schedule.every(360).minutes.do(job,command)
        
        while True:
        	schedule.run_pending()
        	time.sleep(60)
        
    else:
        print('Starting rsync...')
        run_command(command,printout=True)
        return
        


def setup_job(parser):
    jsf = SETTINGS['job_script_filename']
    parser.add_argument('-p','--path',help='Job path (default: %(default)s)',required=False,type=str,
                        default=os.getcwd(),metavar='',dest='path')
    parser.add_argument('-dry','--dry-run',action='store_true',help='Dry run with rsync (default: %(default)s)',default=False,dest='dry_run')
    parser.add_argument('-exc','--exclude',action='append',help='Files to exclude from sync (default: %(default)s)',
                        default=['core.*','WAVECAR'],metavar='',type=str,dest='exclude')
    parser.add_argument('-st','--sync-to-hpc',help='Sync to hpc',required=False,default=False,action='store_true',dest='sync_to_hpc')
    parser.add_argument('-sf','--sync-from-hpc',help='Sync from hpc',required=False,default=False,action='store_true',dest='sync_from_hpc')
    parser.add_argument('-f','--filename',help='Name of bash job script (default: %(default)s)',required=False,default=jsf,type=str,metavar='',dest='filename')
    parser.add_argument('-R','--run',help='Run job on HPC',required=False,default=False,action='store_true',dest='run')
    parser.set_defaults(func=run_job)
    return 
    
def run_job(args):
    j = get_job_from_directory(path=args.path,job_script_filename=args.filename,load_outputs=False)
    
    if args.sync_from_hpc and args.sync_to_hpc:
        raise ValueError("Cannot sync to and from HPC simultaneously")
    elif args.sync_from_hpc:
        j.sync_from_hpc(dry_run=args.dry_run,exclude=args.exclude)
    elif args.sync_to_hpc:
        j.sync_to_hpc(dry_run=args.dry_run,exclude=args.exclude)
        
    if args.run:
        j.run_job(write_input=False,sync=False)
    
    return
    
    







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr  1 15:37:41 2025

@author: villa
"""


import os
import os.path as op
import warnings
import json
import pandas as pd

from monty.json import jsanitize, MontyDecoder
from pymatgen.io.lammps.data import LammpsData
from pymatgen.io.lammps.outputs import parse_lammps_log

from pynter.jobs.core import Job
from pynter.hpc.slurm import JobSettings
from pynter.lammps.inputs import LammpsInput
from pynter import LOCAL_DIR, SETTINGS



def _parse_lammps_log(path,**kwargs):
    try:
        log = parse_lammps_log(op.join(path,'log.lammps'),**kwargs)
        return log
    except:
        warnings.warn('Reading of log.lammps in "%s" failed'%path)
        return False

input_default_filename = SETTINGS['lammps']['input_filename']
data_default_filename = SETTINGS['lammps']['data_filename']
log_default_filename = SETTINGS['lammps']['log_filename']  


class LammpsJob(Job):

        
    def __init__(self,
                 path=None,
                 inputs=None,
                 job_settings=None,
                 outputs=None,
                 job_script_filename=None,
                 name=None,
                 input_filename=input_default_filename,
                 data_filename=data_default_filename,
                 log_filename=log_default_filename,
                 hostname=None,
                 localdir=None,
                 remotedir=None):
        """
        Class to control and organize inputs and outputs of a LAMMPS job.

        Parameters
        ----------
        path : (str)
            Path where job is stored. If None the work dir is used.
        inputs : (dict)
            Dictionary with input data.
        job_settings : (dict)
            Dictionary with job settings. Documentation in JobSettings class in hpc.slurm module
        outputs : (dict), optional
            Dictionary with output data.
        job_script_filename : (str)
            Filename of job script. The default is taken from the key 'filename' in the job_settings in the config file.
        name : (str)
            Name of the job. If none the name is searched in the job script.
        """
        super().__init__(path=path,
                         inputs=inputs,
                         job_settings=job_settings,
                         outputs=outputs,
                         job_script_filename=job_script_filename,
                         name=name,
                         hostname=hostname,
                         localdir=localdir,
                         remotedir=remotedir)
        
        self._input_filename = input_filename
        self._data_filename = data_filename
        self._log_filename = log_filename
        
        if 'data' in self.inputs.keys():
            self._initial_structure = self.inputs['data'].structure
            
    
    def as_dict(self):        
        """
        Get lammpsjob as dictionary. The Vasprun ouput object is not exported.
        
        Parameters
        ----------
        get_band_structure : (bool), optional
            Export BandStructure as dict. The default is False.
            
        Returns:
            Json-serializable dict representation of lammpsjob.
        """                
        d = {"@module": self.__class__.__module__,
             "@class": self.__class__.__name__,
             "path_relative":self.path_relative,
             "inputs": jsanitize(self.inputs),             
             "job_settings": self.job_settings.as_dict(),
             "job_script_filename":self.job_script_filename,
             "name":self.name,
             "hostname":self.hostname,
             "localdir":self.localdir,
             "remotedir":self.remotedir}
        
        d["outputs"] = jsanitize(self.outputs)       
        return d


    def to_json(self,path):
        """
        Save lammpsjob object as json string or file

        Parameters
        ----------
        path : (str), optional
            Path to the destination file.  If None a string is exported.
        get_band_structure : (bool), optional
            Export BandStructure as dict. The default is False.

        Returns
        -------
        d : (str)
            If path is not set a string is returned.

        """
        d = self.as_dict()
        if path:
            with open(path,'w') as file:
                json.dump(d,file)
            return
        else:
            return d.__str__()   

    
    @staticmethod
    def from_dict(d):
        """
        Construct LammpsJob object from python dictionary.
        
        Returns
        -------
        lammpsjob object
        """
        #ensure compatibility with old path format
        if 'path_relative' in d.keys() and d['path_relative']:
            path = LOCAL_DIR + d['path_relative']
        elif 'path' in d.keys():
            path = d['path']
        inputs = {}
        if 'inp' in d['inputs'].keys():
            inputs['inp'] = LammpsInput.from_dict(d['inputs']['inp'])
        if 'data' in d['inputs'].keys():
            inputs['data'] = LammpsData.from_dict(d['inputs']['data'])
        job_settings = JobSettings.from_dict(d['job_settings'])
        job_script_filename = d['job_script_filename']
        name = d['name']
        hostname = d['hostname'] if 'hostname' in d.keys() else None
        localdir = d['localdir'] if 'localdir' in d.keys() else None
        remotedir = d['remotedir'] if 'remotedir' in d.keys() else None
        outputs= MontyDecoder().process_decoded(d['outputs'])
        
        lammpsjob = LammpsJob(path=path,
                          inputs=inputs,
                          job_settings=job_settings,
                          outputs=outputs,
                          job_script_filename=job_script_filename,
                          name=name,
                          hostname=hostname,
                          localdir=localdir,
                          remotedir=remotedir)
        
        return lammpsjob
        
    
    @staticmethod
    def from_directory(
                        path=None,
                        job_script_filename=None,
                        input_filename=input_default_filename,
                        data_filename=data_default_filename,
                        log_filename=log_default_filename,
                        atom_style='atomic',
                        load_outputs=True,
                        hostname=None,
                        localdir=None,
                        remotedir=None,
                        **kwargs):
        """
        Builds LammpsJob object from data stored in a directory.
        Input files are read using the internal LammpsInput class.
        Data files are read using Pymatgen's LammpsData class.
        Output files are read using the parse_lammps_log function from Pymatgen.
        Job settings are read from the job script file.

        Parameters
        ----------
        path : (str)
            Path were job data is stored. If None the current wdir is used. The default is None.
        job_script_filename : (str), optional
            Filename of job script. The default is set in the config file.
        load_outputs : (bool)
            Load job outputs. The default is True.
        kwargs : (dict)
            Arguments to pass to parse_log_lammps.
        Returns
        -------
        lammpsjob object.
        
        """
        path = path if path else os.getcwd()
        inputs = {}
        inputs['inp'] = LammpsInput.from_file(op.join(path,input_filename))
        if os.path.isfile(op.join(path,data_filename)):
            inputs['data'] = LammpsData.from_file(op.join(path,data_filename),atom_style=atom_style)
        outputs = {}
        if load_outputs:
            if op.isfile(op.join(path,log_filename)):
                outputs['log'] = _parse_lammps_log(path,**kwargs)

        job_script_filename = job_script_filename if job_script_filename else SETTINGS['job_script_filename']
        job_settings = JobSettings.from_bash_file(path,filename=job_script_filename)
        
        return LammpsJob(path=path,
                         inputs=inputs,
                         job_settings=job_settings,
                         outputs=outputs,
                         job_script_filename=job_script_filename,
                         input_filename=input_filename,
                         data_filename=data_filename,
                         log_filename=log_filename,
                         hostname=hostname,
                         localdir=localdir,
                         remotedir=remotedir
                         )


    @staticmethod
    def from_json(path_or_string):
        """
        Build LammpsJob object from json file or string.

        Parameters
        ----------
        path_or_string : (str)
            If an existing path to a file is given the object is constructed reading the json file.
            Otherwise it will be read as a string.

        Returns
        -------
        lammpsjob object.

        """
        if op.isfile(path_or_string):
            with open(path_or_string) as file:
                d = json.load(file)
        else:
            d = json.loads(path_or_string)
        return LammpsJob.from_dict(d)


    @property
    def inp(self): # avoid input name in python
        return self.inputs['inp']
    
    @property
    def data(self):
        if 'data' in self.inputs.keys():
            return self.inputs['data']

    @property
    def input_filename(self):
        return self._input_filename

    @property
    def data_filename(self):
        return self._data_filename

    @property
    def log_filename(self):
        return self._log_filename
    
    
    @property
    def formula(self):
        """Complete formula from initial structure (read with Pymatgen)"""
        if self.initial_structure:
            return self.initial_structure.composition.formula
        else:
            return None
    
    @property
    def initial_structure(self):
        if hasattr(self, '_initial_structure'):
            return self._initial_structure
        elif 'data' in self.inputs.keys():
            self._initial_structure = self.inputs['data'].structure
            return self._initial_structure
        else:
            return None      
    
    
    @property
    def log(self):
        if 'log' in self.outputs.keys():
            return self.outputs['log']
        else:
            if not op.exists(op.join(self.path,self.log_filename)):
                warnings.warn(f'"{self.log_filename}" file is not present in Job directory')
            return None

    
    @property
    def is_converged(self):
        if self.outputs:
            if 'convergence' in self.outputs.keys():
                return self.outputs['convergence']
    
    def get_inputs(self,sync=False,atom_style='atomic'):
        if sync:
            self.sync_from_hpc()
        inputs = {}
        inputs['inp'] = LammpsInput.from_file(op.join(self.path,self.input_filename))
        if os.path.isfile(op.join(self.path,self.data_filename)):
            inputs['data'] = LammpsData.from_file(op.join(self.path,self.data_filename),atom_style=atom_style)
        return inputs
    
                
    def get_outputs(self,sync=False,log_filename=log_default_filename,get_output_properties=True,**kwargs):
        """
        Get outputs dictionary from the data stored in the job directory. "log.lammps" is 
        read with Pymatgen

        Parameters
        ----------
        sync : (float)
            Sync data from hpc. The default is False.
        log_filename: (str)
            Name of the lammps log file.
        get_output_properties : (float)
            Parse output properties from log file.
        **kwargs : (dict)
            Arguments for the Vasprun class in pymatgen.
        """
        if sync:
            self.sync_from_hpc()
        self._log_filename = log_filename
        outputs = {}
        if op.isfile(op.join(self.path,log_filename)):
            outputs['log'] = _parse_lammps_log(self.path,**kwargs)

        self.outputs = outputs
        if get_output_properties:
            self.get_output_properties()
        return

        
        

    def get_output_properties(self): #initial idea, to implement further based on specific needs
        self.outputs['convergence'] = self._get_convergence()
        return 
        
    
    def write_input(self):
        self.job_settings.write_bash_file(path=self.path,filename=self.job_settings['filename'])
        self.inp.write_file(filename=op.join(self.path,self.input_filename))
        if self.data:
            self.data.write_file(filename=op.join(self.path,self.data_filename))
    
        

    def _get_convergence(self):
        is_converged = None
        if self.outputs:
            if 'log' in self.outputs.keys():
                if self.log is not None:
                    is_converged = False    
                    if self.log and len(self.log) > 1 and type(self.log[-1]) == pd.DataFrame: # check if last element on the list is a dataframe - maybe revisit this?
                        is_converged = True
        return is_converged
        






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 10 13:30:52 2025

@author: villa
"""

import os
import os.path as op

from pymatgen.io.vasp.inputs import Kpoints, Poscar, Potcar, VaspInput, Incar
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.symmetry.bandstructure import HighSymmKpath

from pynter.hpc.slurm import JobSettings
from pynter.jobs.vasp.vasp_jobs import VaspJob


from pynter import SETTINGS


class DefaultInputs:
    """
    Class to generate VASP input files with default parameters. Use with extreme care.    
    """
    
    def __init__(self,structure=None,sort_structure=True):
        """
        Parameters
        ----------
        structure : Pymatgen Structure object. The default is None. If no Structure is given only default INCAR can be generated.
        sort_structure (bool) : Sort structure to avoid POSCAR/POTCAR inconsistencies. The default is True.
        """
        if structure and sort_structure:
            structure.sort()
        self._structure = structure if structure else None 
        defaults = SETTINGS['vasp']
        
        self.default_potcar_symbols =  defaults['default_potcar_symbols']
        if self.structure:
            self.potcar_symbols = [self.default_potcar_symbols[el.symbol] for el in self.structure.composition.elements]
        self.incar_default_flags = defaults['incar_default']
        
    @property
    def structure(self):
        return self._structure
    
    @structure.setter
    def structure(self,structure):
        self._structure = structure

    def get_incar_default(self, xc='PBE', ldauu=None, aexx=0.25):
        """
        Get dictionary with default INCAR parameters

        Parameters
        ----------
        xc : (str), optional
            Which functional to use('PBE','PBE+U','LDA','LDA+U',HSE06' available). The default is 'PBE'.
        ldauu : (List), optional
            List of integers for U parameter to be given in order as required by VASP. The default is None. If None no 'LDAUU' flag is written.
        aexx : (Int), optional
            Value of Hartree-Fock contribution for HSE calculation. The default is 0.25.

        Returns
        -------
        incar_default_flags : Dict
            Dictionary with default INCAR parameters.

        """
        system = self.structure.composition.reduced_formula if self.structure else 'No system info'
          
        incar_default_flags = self.incar_default_flags              
        incar_default_flags["SYSTEM"] = system                   
                     
        if xc == 'PBE' or xc == 'LDA':
            incar_default_flags.update({
                    "ISYM":2})
            
        if xc == 'PBE+U' or xc == 'LDA+U':
            incar_default_flags.update({
                    "ISYM":2,
                    "LDAU" : ".TRUE.",
                    "LDAUTYPE": 2,
                    "LDAUPRINT": 2,
                })
            if ldauu:
                incar_default_flags.update({"LDAUU": ' '.join([str(u) for u in ldauu])}) # if LDAUU is not present VASP will put all zeros
            
        if xc == 'HSE06':
            incar_default_flags.update({
                    "LHFCALC" : ".TRUE.",
                    "HFSCREEN": 0.2,
                    "PRECFOCK": "Fast",
                    "AEXX": aexx,
                    "ISYM":3
                    })
    
        return incar_default_flags


    def get_vasp_input(self, xc='PBE', ldauu=None, aexx=0.25, kppa=1000,
                       potcar_symbols=None, potcar_functional='PBE'):
        """
        Get pymatgen VaspInput object with complete set of default inputs.

        Parameters
        ----------
        xc : (str), optional
            Which functional to use('PBE','PBE+U','HSE06' available). The default is 'PBE'.
        ldauu : (List), optional
            List of integers for U parameter to be given in order as required by VASP. The default is None. If None no 'LDAUU' flag is written.
        aexx : (Int), optional
            Value of Hartree-Fock contribution for HSE calculation. The default is 0.25.
        kppa : (Int), optional
            Value of k-points density per atom to generate automatic k-mesh as done by Pymatgen. The default is 1000.
        potcar_symbols : List, optional
            List of strings with symbols of POTCARs. The default is None. If None the default symbols from the Materials Project \n
            database are used for every specie in the given Structure object.
        potcar_functional : (str), optional
            Functional to be used, as defined by Pymatgen Potcar class. The default is 'PBE'.

        Returns
        -------
        Pymatgen VaspInput object.

        """
        
        incar = Incar(self.get_incar_default(xc=xc,ldauu=ldauu,aexx=aexx))
        kpoints = self.get_kpoints_default(kppa=kppa)
        poscar = self.get_poscar()
        potcar = self.get_potcar(potcar_symbols=potcar_symbols,potcar_functional=potcar_functional)
        
        return VaspInput(incar,kpoints,poscar,potcar)
        
        
    def get_kpoints_default(self,kppa=1000):
        """
        Get Pymatgen Kpoints object with automatic Gamma centered k-mesh generation from given Structure object.

        Parameters
        ----------
        kppa : (Int), optional
            Value of k-points density per atom to generate automatic k-mesh as done by Pymatgen. The default is 1000.

        Returns
        -------
        Pymatgen Kpoint object.
        """
        
        if self.structure:
            structure = self.structure
        else:
            raise ValueError('Structure object needs to be given to obtain pymatgen automatic kpoint mesh')
         
        return Kpoints().automatic_gamma_density(structure,kppa)
    
  
    def get_kpoints_bs_default(self,divisions=10,hybrid_mode=False,kppa=1000):
        """
        Get Pymatgen Kpoints object for band structure calculations.
        The default high symmetry path for PBE from Pymatgen class HighSymmKpath is obtained from 
        the input Structure with 10 points between high symm k-points.
        For Hybrid BS the calculations need to be selfconsistent, the KPOINTS file is a list of single kpoints.
        The irriducible kpoints with weight != 0 are obtained with Pymatgen's SpaceGroupAnalyzer.
        The additional points with weight = 0 on the high symm path are obtained with the HighSymmPath class
        like for the PBE case.'

        Parameters
        ----------
        divisions : int, optional
            Number of points in between high symmetry points. The default is 10.
        hybrid_mode : (bool), optional
            Kpoints for hybrid calculation. The default is False.
        kppa : (Int), optional
            Value of k-points density per atom to generate automatic k-mesh as done by Pymatgen. The default is 1000.

        Returns
        -------
        Pymatgen Kpoint object.
        """
        if hybrid_mode: 
            # code copy/pasted (slightly adjusted) from pymatgen.io.vasp.sets.MPHSEBSSet
            # The whole input set was not needed but only the kpoints generation in the kpoints @property
            kpts = []
            weights = []  
            all_labels = []  
            structure = self.structure
    
            # for both modes, include the Uniform mesh w/standard weights
            grid = Kpoints().automatic_gamma_density(structure, kppa).kpts
            ir_kpts = SpacegroupAnalyzer(structure, symprec=0.1).get_ir_reciprocal_mesh(grid[0])
            for k in ir_kpts:
                kpts.append(k[0])
                weights.append(int(k[1]))
                all_labels.append('')
       
            #add the symmetry lines w/zero weight
            kpath = HighSymmKpath(structure)
            frac_k_points, labels = kpath.get_kpoints(line_density=divisions,coords_are_cartesian=False)

            for k, f in enumerate(frac_k_points):
                kpts.append(f)
                weights.append(0.0)
                all_labels.append(labels[k])
    
            comment = "HSE run along symmetry lines"   
            kpoints =  Kpoints(comment=comment,style=Kpoints.supported_modes.Reciprocal,num_kpts=len(kpts),
                kpts=kpts,kpts_weights=weights,labels=all_labels)

        else:
            kpoints = Kpoints().automatic_linemode(10,HighSymmKpath(self.structure))
    
        return kpoints
    
    
    def get_poscar(self):
        """
        Get Pymatgen Poscar object from given Structure object.        

        Returns
        -------
        Pymatgen Poscar object
        """
        
        if self.structure:
            pass
        else:
            raise ValueError('Structure object needs to be given to get Poscar onject')
        
        return Poscar(self.structure)
        
     
    def get_potcar(self,potcar_symbols=None,potcar_functional='PBE'):
        """
        Get Pymatgen Potcar object. The PMG_VASP_PSP directory has to be defined in .pmgrc.yaml file as required by Pymatgen.
        Parameters
        ----------
        potcar_symbols : List, optional
            List of strings with symbols of POTCARs. The default is None. If None the default symbols from the Materials Project \n
            database are used for every specie in the given Structure object.
        potcar_functional : (str), optional
            Functional to be used, as defined by Pymatgen Potcar class. The default is 'PBE'.

        Returns
        -------
        Pymatgen Potcar object.
        """
                
        if potcar_symbols:
            potcar = Potcar(symbols=potcar_symbols,functional=potcar_functional)
            self.potcar_symbols = potcar_symbols
        else:
            if self.structure:
                potcar_symbols = self.potcar_symbols
            else:
                raise ValueError('potcar symbols or Structure object need to be given to get Potcar onject')
            potcar = Potcar(symbols=potcar_symbols,functional=potcar_functional)
        
        return potcar
            

class DefaultJobSettings:
    
    def __init__(self,
                 vasp_exe=None,
                 vasp_sbatch=None,
                 modules=None,
                 lines_before_srun=None,
                 lines_after_srun=None):
        """
        Class to handle VASP specific part of sbatch submission script.
        Args that are not provided are taken from vasp.yml file in ~/.pynter.

        Parameters
        ----------
        vasp_exe : (str)
            Path to vasp executable. 
        vasp_sbatch : (dict)
            Sbatch dictionary tailored for vasp.
        modules : (list)
            Name of modules to load with "module load".
        lines_before_srun : (list)
            Script lines to be placed before the "srun" command.
        lines_after_srun : (list)
            Script lines to be placed before the "srun" command.
        """
        
        default_settings = SETTINGS['vasp']['job_settings_default']
        self.vasp_exe = vasp_exe or default_settings['vasp_exe']
        self.vasp_sbatch = vasp_sbatch or default_settings['vasp_sbatch']
        self.modules = modules or default_settings['modules']
        self.lines_before_srun = lines_before_srun or default_settings['lines_before_srun']
        self.lines_after_srun = lines_after_srun or default_settings['lines_after_srun']
    
    @property
    def vasp_script_lines(self):
        if self.modules:
            lines = self.get_modules_lines()
        else:
            lines = []
        if self.lines_before_srun:
            lines += self.lines_before_srun
        if self.vasp_exe:
            lines += [self.get_srun_line()]
        if self.lines_after_srun:
            lines += self.lines_after_srun
        return lines
        
    def get_modules_lines(self):
        if self.modules:
            return [f'module load {module}' for module in self.modules] 

    def get_srun_line(self):
        return 'srun %s' %self.vasp_exe

    def get_updated_job_settings(self,job_settings):
        """
        Get a copy of a JobSettings object updated with VASP specific settings
        """
        job_settings = job_settings.copy()
        for k,v in self.vasp_sbatch.items():
            job_settings[k] = v
                                
        lines = self.vasp_script_lines
        for line_to_add in lines:
            if all([line_to_add not in line for line in job_settings['script_lines']]):
                job_settings['script_lines'].append(line_to_add)

        return job_settings
    
    

class InputSets:
    
    def __init__(self,
                 path=None,
                 vaspinput=None,
                 structure=None,
                 incar_settings=None,
                 kpoints=None,
                 potcar=None,
                 job_settings=None,
                 name=None,
                 add_parent_folder=False,
                 **kwargs):
        """
        Parameters
        ----------
        path : (str)
            Main path for the scheme. If None the current work dir is used. The default is None.
        vaspinput : (Pymatgen VaspInput object), optional
            Set of VASP inputs, the default is None. if provided the other inputs of the class (structure,incar_settings,kpoints,potcar) are not needed 
        structure : (Pymatgen Structure object), optional
            Pymatgen Structure object.
        incar_settings : (Dict), optional
            Dictionary with incar flags. The default is None. If None the default settings for PBE functional from the DefaultInputs class are used.
        kpoints : (Pymatgen Kpoints object), optional
            Pymatgen Kpoints object. The default is None. If None the default settings from the DefaultInputs class are used.
        potcar : (Pymatgen kpoints object), optional
            Pymatgen kpoints object. The default is None. If None the default settings from the DefaultInputs class are used.
        job_settings : (Dict), optional
            Dictionary with job settings to create job script, parameters are defined in JobSettings class function. The default is None.\n
            If job_settings is None, the 'job-name' key will be added, the value is the 'name' argument if provided, if 'name' arg is \n
            None the value will be: 'no_name'.
        name : (str), optional
            Name for the system to set up scheme for. The default is None.
        add_parent_folder : (bool), optional
            Add folder to the path names like the name of the InputSets. Default is False.
        """
        self.path = op.abspath(path) if path else os.getcwd()
        if vaspinput:
            structure = vaspinput['POSCAR'].structure
            incar_settings = Incar(vaspinput['INCAR'].copy())
            kpoints = vaspinput['KPOINTS']
            potcar = vaspinput['POTCAR']

        self.structure = structure if structure else None
        if structure:
            self.incar_settings = incar_settings if incar_settings else DefaultInputs(self.structure).get_incar_default()
            self.kpoints = kpoints if kpoints else DefaultInputs(self.structure).get_kpoints_default()
            self.potcar = potcar if potcar else DefaultInputs(self.structure).get_potcar()
            if job_settings:
                self.job_settings = JobSettings(**job_settings)
            elif name:
                self.job_settings = JobSettings(sbatch={'job-name':name})
            else:
                self.job_settings = JobSettings(sbatch={'job-name':'no_name'})
            self.name = name if name != None else self.job_settings['job-name'] 
            
            if self.name: # this return false if name is '', the previuos line considers only if name is None
                self.job_settings['job-name'] = self.name
            
         #   if all(['srun' not in line for line in self.job_settings['script_lines']]):
            self.job_settings = DefaultJobSettings(**kwargs).get_updated_job_settings(self.job_settings)    
            
        else:
            raise ValueError('You need to provide Structure, either as Poscar object in VaspInput or in "structure" arg')

        if add_parent_folder:
            self.path = op.join(self.path,self.name)

    def __str__(self):
        printout = 'System name:"%s" \n' %self.name
        printout += 'STRUCTURE:\n'
        printout += self.structure.__str__() + '\n'
        printout += 'INCAR:\n'
        printout += '%s \n' %str(self.incar_settings)
        printout += 'KPOINTS:\n'
        printout += self.kpoints.__str__() + '\n'
        printout += 'POTCAR symbols:\n'
        printout += '%s \n' %(', '.join(self.potcar.symbols))
        printout += 'Job settings:\n'
        printout += '%s' %(str(self.job_settings))
        return printout
    
    def __repr__(self):
        return self.__str__()


    def get_vaspjob(self,add_to_job_name=None,add_to_job_path=None):
        """
        Generate VaspJob object from the input settings of the class.

        Parameters
        ----------
        add_to_job_name : (str), optional
            String to be added to 'name' key in job_settings dictionary and to name attribute of VaspJob.
        add_to_job_path : (str), optional
            String to be added to self.path. The complete path will be input in 'path' arg of VaspJob.

        Returns
        -------
        VaspJob object
        """        
        incar_settings = self.incar_settings.copy()
        job_settings = self.job_settings.copy()
        
        incar = Incar(incar_settings)
        kpoints = self.kpoints
        poscar = Poscar(self.structure)
        potcar = self.potcar
        vaspinput = VaspInput(incar,kpoints,poscar,potcar)
        
        if add_to_job_name:
            job_settings['job-name'] = '_'.join([job_settings['job-name'],add_to_job_name])        
            jobname = '_'.join([self.name,add_to_job_name])
        else:
            jobname = self.name
            
        if add_to_job_path:
            jobpath = op.join(self.path,add_to_job_path)
        else:
            jobpath = self.path
            
        job_script_filename = job_settings['filename'] if 'filename' in job_settings.keys() else None
        vaspjob = VaspJob(path=jobpath,inputs=vaspinput,job_settings=job_settings,job_script_filename=job_script_filename,name=jobname)
        
        return vaspjob  


    def hse_bs(self,kpoints_bs=None,add_to_job_name='HSE-BS',add_to_job_path='HSE-BS',**kwargs):
        """
        Set up band structure calculation with HSE
        
        Parameters
        ----------
        kpoints_bs : (Pymatgen Kpoints object) , optional
            Pymatgen Kpoints object for the path in BS calculation. The default is None. If None the default high symmetry path \n
            from Pymatgen class HighSymmKpath is obtained from the input Structure with 10 points between high symm k-points.
        kwargs : (dict), optional
            Kwargs for get_kpoints_bs_default in DefaultInputs
        """
        vaspjob = self.hse_scf(add_to_job_name,add_to_job_path)
        if kpoints_bs:
            kpoints = kpoints_bs
        else:
            kpoints = DefaultInputs(self.structure).get_kpoints_bs_default(hybrid_mode=True,**kwargs)
        vaspjob.inputs['KPOINTS'] = kpoints
        return vaspjob

    def hse_dos(self,kmesh=3,add_to_job_name='HSE-DOS',add_to_job_path='HSE-DOS'):
        """
        Set up DOS calculation with HSE

        Parameters
        ----------
        kmesh : (Int), optional
            Multiplier for coefficients of the k-mesh for the DOS with respect to first step SCF calculation. The default is 3.
        """
        vaspjob = self.hse_scf(add_to_job_name,add_to_job_path)
        vaspjob.incar['NEDOS'] = 2000
        vaspjob.incar['ISMEAR'] = -5        
        # multiply by 3 coeff of k-mesh
        kpoints = vaspjob.inputs['KPOINTS']
        style, kpts, kpts_shift = kpoints.style, kpoints.kpts, kpoints.kpts_shift
        kpts_new = []
        kpts_new.append([k*kmesh for k in kpts[0]])
        kpoints = Kpoints(style=style,kpts=kpts_new,kpts_shift=kpts_shift)
        vaspjob.inputs['KPOINTS'] = kpoints
        return vaspjob        

    def hse_ionic_rel(self,add_to_job_name='HSE-rel',add_to_job_path='HSE-rel'):
        """
        Set up ionic relaxation with HSE
        """
        vaspjob = self.pbe_ionic_rel(add_to_job_name,add_to_job_path)
        vaspjob.incar['LHFCALC'] = '.TRUE.'
        vaspjob.incar['ISYM'] = 3
        vaspjob.job_settings['array_size'] = 7
        return vaspjob

    def hse_ionic_vol_rel(self,add_to_job_name='HSE-rel',add_to_job_path='HSE-rel'):
        """
        Relaxation of atomic positions and cell volume with HSE
        """
        vaspjob = self.hse_ionic_rel(add_to_job_name,add_to_job_path)
        vaspjob.incar['ISIF'] = 3        
        return vaspjob

    def hse_ionic_rel_gamma(self,add_to_job_name='HSE-rel-Gamma',add_to_job_path='HSE-rel-gamma'):
        """
        Set up HSE ionic relaxation only in gamma point 
        """
        vaspjob = self.hse_ionic_rel(add_to_job_name,add_to_job_path)
        vaspjob.inputs['KPOINTS'] = Kpoints().gamma_automatic(kpts=(1,1,1))         
        return vaspjob 

    def hse_scf(self,add_to_job_name='HSE-SCF',add_to_job_path='HSE-SCF'):
        """
        Set up SCF calculation with HSE
        """
        vaspjob = self.pbe_scf(add_to_job_name,add_to_job_path)
        vaspjob.incar['LHFCALC'] = '.TRUE.'
        vaspjob.incar['ISYM'] = 3  
        return vaspjob

    def hse_scf_gamma(self,add_to_job_name='HSE-SCF-Gamma',add_to_job_path='HSE-SCF-Gamma'):
        """
        Set up HSE-SCF calculation only in gamma point 
        """
        vaspjob = self.hse_scf(add_to_job_name,add_to_job_path)
        vaspjob.inputs['KPOINTS'] = Kpoints().gamma_automatic(kpts=(1,1,1))        
        return vaspjob 

    def hse_standard(self,add_to_job_name='HSE',add_to_job_path='HSE'):
        """
        Set for standard HSE caculation
        """
        vaspjob = self.get_vaspjob(add_to_job_name,add_to_job_path)
        vaspjob.incar['LHFCALC'] = '.TRUE.'
        vaspjob.incar['ISYM'] = 3 
        return vaspjob


    def hubbard(self,ldauu_dict,add_to_job_name,add_to_job_path):
        """
        Set for standard PBE+U calculation. The dict {Element:U_value} is used to construct the INCAR
        """
        vaspjob = self.get_vaspjob(add_to_job_name,add_to_job_path)
        vaspjob.incar['LDAU'] = '.TRUE.'
        vaspjob.incar['LDAUTYPE'] = 2
        vaspjob.incar['LDAUPRINT'] = 2
        vaspjob.incar['LDAUU'] = ' '.join([str(ldauu_dict[el]) for el in ldauu_dict])
        return vaspjob


    def pbe_bs(self,kpoints_bs=None,add_to_job_name='PBE-BS',add_to_job_path='PBE-BS',**kwargs):
        """
        Set up band structure calculation with PBE
        
        Parameters
        ----------
        kpoints_bs : (Pymatgen Kpoints object) , optional
            Pymatgen Kpoints object for the path in BS calculation. The default is None. If None the default high symmetry path \n
            from Pymatgen class HighSymmKpath is obtained from the input Structure with 10 points between high symm k-points.
        kwargs : (dict), optional
            Kwargs for get_kpoints_bs_default in DefaultInputs
        """
        vaspjob = self.pbe_scf(add_to_job_name,add_to_job_path)
        vaspjob.incar['ICHARG'] = 11
        vaspjob.incar['LORBIT'] = 11
        vaspjob.incar['ISMEAR'] = 0
        vaspjob.incar['NEDOS'] = 2000
        if kpoints_bs:
            kpoints = kpoints_bs
        else:
            kpoints = DefaultInputs(self.structure).get_kpoints_bs_default(**kwargs)
        vaspjob.inputs['KPOINTS'] = kpoints
        return vaspjob

    def pbe_dos(self,kmesh=3,add_to_job_name='PBE-DOS',add_to_job_path='PBE-DOS'):
        """
        Set up DOS calculation with PBE

        Parameters
        ----------
        kmesh : (Int), optional
            Multiplier for coefficients of the k-mesh for the DOS with respect to first step SCF calculation. The default is 3.
        """
        vaspjob = self.pbe_scf(add_to_job_name,add_to_job_path)
        vaspjob.incar['NEDOS'] = 2000
        vaspjob.incar['ISMEAR'] = -5        
        # multiply by 3 coeff of k-mesh
        kpoints = vaspjob.inputs['KPOINTS']
        style, kpts, kpts_shift = kpoints.style, kpoints.kpts, kpoints.kpts_shift
        kpts_new = []
        kpts_new.append([k*kmesh for k in kpts[0]])
        kpoints = Kpoints(style=style,kpts=kpts_new,kpts_shift=kpts_shift)
        vaspjob.inputs['KPOINTS'] = kpoints
        return vaspjob
 
    def pbe_ionic_rel(self,add_to_job_name='PBE-rel',add_to_job_path='PBE-rel'):
        """
        Relaxation of atomic positions with PBE
        """
        vaspjob = self.pbe_scf(add_to_job_name,add_to_job_path)
        vaspjob.incar['NSW'] = 100
        vaspjob.incar['EDIFF'] = 1e-05
        vaspjob.incar['ISIF'] = 2        
        return vaspjob 

    def pbe_ionic_rel_gamma(self,add_to_job_name='PBE-rel-Gamma',add_to_job_path='PBE-rel-gamma'):
        """
        Set up PBE ionic relaxation only in gamma point 
        """
        vaspjob = self.pbe_ionic_rel(add_to_job_name,add_to_job_path)
        vaspjob.inputs['KPOINTS'] = Kpoints().gamma_automatic(kpts=(1,1,1))        
        return vaspjob 

    def pbe_ionic_vol_rel(self,add_to_job_name='PBE-rel-vol',add_to_job_path='PBE-rel-vol'):
        """
        Relaxation of atomic positions and cell volume with PBE
        """
        vaspjob = self.pbe_ionic_rel(add_to_job_name,add_to_job_path)
        vaspjob.incar['ISIF'] = 3        
        return vaspjob 
                
    def pbe_scf(self,add_to_job_name='PBE-SCF',add_to_job_path='PBE-SCF'):
        """
        Set up standard input for electronic minimization with PBE
        """        
        vaspjob = self.get_vaspjob(add_to_job_name,add_to_job_path)
        vaspjob.incar['NSW'] = 0
        vaspjob.incar['EDIFF'] = 1e-06        
        return vaspjob      

    def pbe_scf_gamma(self,add_to_job_name='PBE-SCF-Gamma',add_to_job_path='PBE-SCF-Gamma'):
        """
        Set up PBE-SCF calculation only in gamma point 
        """
        vaspjob = self.pbe_scf(add_to_job_name,add_to_job_path)
        vaspjob.inputs['KPOINTS'] = Kpoints().gamma_automatic(kpts=(1,1,1))         
        return vaspjob 


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Aug 28 14:23:05 2023

@author: villa
"""

import os
import os.path as op
import warnings
import copy
import re

from monty.json import MSONable

from pynter import SETTINGS
from pynter.tools.utils import grep_list


def read_possible_sbatch_arguments():
    """
    Read possible sbatch arguments from txt file and map (when present) their short versions 
    (-<letter>) with their long versions (--<word>).

    Returns
    -------
    arguments : (list)
        List containing all possible arguments.
    arguments_legend : (dict)
        Dictionary containing argument mapping.
    """
    path = os.path.abspath(__file__).strip(os.path.basename(__file__))
    filename = 'sbatch_arguments.txt'
    with open(os.path.join(path,filename),'r') as file:
        lines = file.readlines()
  
    arguments = []
    arguments_legend = {}
    for line in lines:
        line = line.strip('\n')
        elements = line.split(',')
        arg = elements[0]
        arguments.append(arg)
        if len(elements) > 1:
            arg_short = elements[1]
            arguments_legend.update({arg_short:arg})

    return arguments, arguments_legend


class Sbatch(dict,MSONable):
    
    def __init__(self,load_defaults=True,**kwargs):
        """
        Object to handle sbatch commands in sbatch. Behaves like a dictionary.

        Parameters
        ----------
        **kwargs : 
            All possible sbatch arguments.
        """
        super().__init__()
        self._arguments, self._arguments_legend = read_possible_sbatch_arguments()
        
        defaults = SETTINGS['sbatch']
        
        settings = {}
        if load_defaults:
            for key,value in defaults.items():
                settings[key] = value
        
        for key,value in kwargs.items():
            if key in self.arguments_legend.keys():
                key = self.arguments_legend[key]
            settings[key] = value
        
        settings = dict(sorted(settings.items(), key=lambda item: item[0]))
        self.update(settings)
        
    
    def __setitem__(self,key,value):
        if key not in self.arguments:
            warnings.warn(f'"{key}" is not listed as possible sbatch argument')
        else:
            super().__setitem__(key,value)
        return
    
    def as_dict(self):
        d = dict(self)
        d["@module"] = type(self).__module__
        d["@class"] = type(self).__name__
        return d        
    
    def copy(self):
        return copy.deepcopy(self)
    
    def update(self, other=None, **kwargs):
        if other is not None:
            if hasattr(other, "items"):
                for key,value in other.items():
                    self.__setitem__(key, value)
            else:
                for key, value in other:
                    self.__setitem__(key, value)
        return
    
    @classmethod
    def from_dict(cls,d):
        return cls(**{k: v for k, v in d.items() if k not in ["@module", "@class"]})
    
    
    @property
    def arguments(self):
        """
        List of all possible sbatch arguments
        """
        return self._arguments
    
    @property
    def arguments_legend(self):
        """
        Dictionary which maps the short version (-<letter>) of the
        arguments with the long one (--<word>)
        """
        return self._arguments_legend
    
    
    @staticmethod
    def from_bash_script(input_string,load_defaults=False):
        """
        Initialize object from a bash script
        """
        lines = input_string.split('\n')
        kwargs = {}
        string = '#SBATCH'
        target_lines = grep_list(string,lines)
        for line in target_lines:  
            words = line.split(' ')
            if '-' in line and '--' not in line:
                if len(words) > 2:
                    pair = words[1].strip('-') , words[2]
                else:
                    pair = words[1].strip('-')
            elif '--' in line:
                pair =  words[1].strip('--').split('=')
            key = pair[0]
            value = pair[1] if len(pair) > 1 else ''
            if value.isdigit():
                value = int(value)
            kwargs.update({key:value})
            
        return Sbatch(**kwargs,load_defaults=load_defaults)
     

    def get_bash_script_lines(self):
        """
        Get list of lines containing #SBATCH arguments for bash script
        """
        lines = []
        lines.append('#!/bin/sh\n')
        for key,value in self.items():
            if value is not None:
                printed_value = f'={value}' if value else '' 
                lines.append(f'#SBATCH --{key}{printed_value}')
        return lines
                


class JobSettings(dict,MSONable):
    
    def __init__(self,load_sbatch_defaults=True,sbatch={},filename=None,script_lines=[]):
        """
        Handles settings for creation of bash scripts for Job submissions (sbatch). 
        Behaves like a dictionary. The sbatch keys names are consistent with the 
        arguments of the sbatch command on the slurm system. Additional text in the
        script is included as a list of lines. 

        Parameters
        ----------
        load_sbatch_defaults : (bool)
            Load default sbatch arguments from configuration file.
        sbatch : (dict or Sbatch)
            Dictionary with sbatch commands.
        filename : (str)
            Default name for bash file. If None the default is taken from configuration file.
        script_lines : (list)
            Additional lines for the bash script.
        """
        
        super().__init__()
        
        if sbatch:
            if type(sbatch) is dict:
                sbatch = Sbatch(load_defaults=load_sbatch_defaults,**sbatch)
            elif sbatch.__class__.__name__ == 'Sbatch':
                sbatch = sbatch
            else:
                raise ValueError('sbatch settings must be provided either as dictionary or as Sbatch object')                
        elif not sbatch:
            sbatch = Sbatch(load_defaults=load_sbatch_defaults,**sbatch)
        
        settings = {} 
        settings['sbatch'] = sbatch
        settings['filename'] = filename or SETTINGS['job_script_filename']
        settings['script_lines'] = script_lines
        self.update(settings)
        

    def __repr__(self):
        return self.get_bash_script()
    
    def __print__(self):
        return self.__repr__()


    def __getitem__(self,key):
        if key not in self.keys():
            return self['sbatch'][key]
        else:
            return super().__getitem__(key)
    
    def __setitem__(self,key,value):
        if key not in ['filename','script_lines']:
            self['sbatch'].__setitem__(key, value)
        elif key != 'sbatch':
            super().__setitem__(key,value)
        return

    def __deepcopy__(self, memo):
        new_obj = type(self)(load_sbatch_defaults=False,
                              sbatch=copy.deepcopy(self['sbatch'], memo), 
                              filename=self['filename'],
                              script_lines=self['script_lines'][:])
        return new_obj
        
    def copy(self):        
        return copy.deepcopy(self,{})

    
    def as_dict(self):
        d = dict(self)
        d["@module"] = type(self).__module__
        d["@class"] = type(self).__name__
        return d       
    
    @classmethod
    def from_dict(cls,d):
        return cls(**{k: v for k, v in d.items() if k not in ["@module", "@class"]})


    @staticmethod
    def from_bash_file(path,filename='job.sh'):
        """
        Create JobSettings object from file. Cannot read added lines in header and body
        """
        file = op.join(path,filename)
        with open(file) as f:
            input_string = f.read()
        job_settings = JobSettings.from_bash_script(input_string)
        job_settings['filename'] = filename
        return job_settings

    @staticmethod
    def from_bash_script(input_string):
        """
        Initialize object from a bash script
        """
        sbatch = Sbatch.from_bash_script(input_string)
        lines = input_string.split('\n')
        script_lines = []
        for line in lines:
            if all(target_string not in line for target_string in ['#SBATCH','#!']):
                   if line:
                       script_lines.append(line)
         
        return JobSettings(load_sbatch_defaults=False,sbatch=sbatch,script_lines=script_lines)


    def get_bash_script(self):
        lines = self['sbatch'].get_bash_script_lines()
        lines.append('')
        lines += self['script_lines'] 
        return '\n'.join(lines)
    
    
    def replace_line(self,old_string,new_string):
        for index, line in enumerate(self['script_lines']):
            if old_string in line:
                new_line = line.replace(old_string, new_string)
                self['script_lines'][index] = new_line
                
        return

    
    def write_bash_file(self,path=None,filename=None):
        """
        Write job script file.
        
        Parameters
        ----------
        path : (str), optional
            Path to write job script to. The default is None. If None work dir is used.
        filename : (str), optional
            Filename. If None self['filename'] is used. The default is None.
        """
        if path:
            if not os.path.exists(path):
                os.makedirs(path)
        filename = filename if filename else self['filename']
        complete_path = os.path.join(path,self['filename']) if path else filename      
        with open(complete_path,'w') as f:
            f.write(self.get_bash_script())        
        return





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May  9 14:54:13 2025

@author: villa
"""

import warnings
from monty.json import MSONable


class LammpsInput(MSONable):
    
    def __init__(self,lines):
        self.lines = lines
    
    def __repr__(self):
        return self.get_string()
    
    def __print__(self):
        return self.__repr__()
    
    def __eq__(self,other):
        self_lines = [line for line in self.lines if line] # remove empty lines
        other_lines = [line for line in other.lines if line] 
        return self_lines == other_lines
    
    def copy(self):
        return LammpsInput(self.lines.copy())
    
    @staticmethod
    def from_string(string,remove_empty_lines=False):
        """
        Get LammpsInput object from string
        """
        if remove_empty_lines:
            string = "\n".join(line for line in string.splitlines() if line.strip())
        return LammpsInput(string.split('\n'))
    
    @staticmethod
    def from_file(filename):
        """
        Get LammpsInput object from file
        """
        with open(filename,'r') as file:
            lines = []
            for line in file.readlines():
                lines.append(line.strip('\n'))
        return LammpsInput(lines)
                    
    
    def get_command(self,command):
        """
        Get line containing a specific command
        """
        return self.get_command_and_indexes(command)[0]

    
    def get_command_and_indexes(self,command):
        """
        Get line containing a specific command and relative index
        """
        target_lines = []
        indexes = []
        for index,line in enumerate(self.lines):
            if command in line:
                target_lines.append(line)
                indexes.append(index)
        if len(target_lines) == 0:
            warnings.warn('No line containing command found')
        return target_lines, indexes
    
    def get_string(self):
        """
        Get LAMMPS input string
        """
        return '\n'.join(self.lines)
        
        
    def insert_line(self,line,indexes=None,previous_command=None):
        """
        Insert line at a specific index or after a line containing a target command.

        Parameters
        ----------
        line : (str)
            Line to insert.
        indexes : (int)
            List index in which to insert the line.
        previous_command : (str)
            Insert the line after a specific command. This arg and the indexes
            arg are mutually exclusive, indexes has the priority.
        """
        if indexes:
            pass
        elif previous_command:
            _, indexes = self.get_command_and_indexes(previous_command)
            if len(indexes) > 1:
                warnings.warn('More than one target line found, ')
        else:
            raise ValueError('either indexes or previous command line need to be provided')
        for idx in indexes:
            self.lines.insert(idx+1,line)
        return
    
    
    def set_command(self,command,new_line):
        """
        Change a line containing a specific command.

        Parameters
        ----------
        command : (str)
            Target command to reset..
        new_line : (str)
            New line to insert.
        """
        _, indexes = self.get_command_and_indexes(command)
        for idx in indexes:
            self.lines[idx] = new_line
        return 
            
    
    def write_file(self,filename):
        """
        Write LAMMPS input file
        """
        with open(filename,'w') as file:
            file.writelines('\n'.join(self.lines))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 30 14:45:20 2023

@author: villa
"""

from pynter.hpc.slurm import Sbatch, JobSettings


def setup_job_script(subparsers):
    
    parser_sub = subparsers.add_parser('job-script',help='Create job script')
    parser_sub = parse_job_script_args(parser_sub)
    parser_sub.set_defaults(func=write_job_script)
    return
    

def parse_job_script_args(parser):
    shdef = JobSettings()
    parser.add_argument('--filename',help='File name (default: %(default)s)',required=False,default=shdef['filename'],type=str,metavar='',dest='filename')
    parser.add_argument('--lines',action='append',help='Add line to script',required=False,default=shdef['script_lines'],type=str,metavar='',dest='script_lines')

    sbatch = Sbatch()
    for key in sbatch.arguments:
        if key in sbatch.keys():
            parser.add_argument(f'--{key}',required=False,type=str,default=sbatch[key],metavar='',dest=key)
        else:
            parser.add_argument(f'--{key}',required=False,type=str,metavar='',dest=key)

    return parser 


def write_job_script(args):
    sbatch = vars(args).copy()
    if 'func' in sbatch.keys():
        del sbatch['func']
    del sbatch['filename']
    del sbatch['script_lines']

    JobSettings(sbatch=sbatch,filename=args.filename,script_lines=args.script_lines).write_bash_file()
    return
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 19 14:52:51 2020

@author: villa
"""

from pymatgen.core.periodic_table import Element

from pynter.vasp.schemes.core import InputSets

  
class RelaxationSchemes(InputSets):
    """
    Class to generate and write input files for different calculation schemes in VASP
    """

    def hse_relaxation(self,scheme_name='HSE-rel'):
        """
        Generates calculation scheme for ionic relaxation for HSE. Steps: \n
            '1-PBE-SCF': Electronic SCF with PBE \n
            '2-PBE-OPT': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with PBE \n
            '3-HSE-SCF': Electronic SCF with HSE \n
            '4-HSE-OPT': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with HSE
            
        Returns:
            List of VaspJob objects
        """        
        stepnames = ['1-PBE-SCF','2-PBE-OPT','3-HSE-SCF','4-HSE-OPT']
        jobs = []
        
        sn = 1 #set step number 1
        vaspjob = self.pbe_scf(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 2
        vaspjob = self.pbe_ionic_rel(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 3 
        vaspjob = self.hse_scf(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 4
        vaspjob = self.hse_ionic_rel(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)

        return jobs
    
    
    def hse_relaxation_gamma(self,scheme_name='HSE-rel-gamma'):
        """
        Generates calculation scheme for structure relaxations for HSE with more intermediate steps. Steps: \n
            '1-PBE-SCF-Gamma': Electronic SCF with PBE only in Gamma point \n
            '2-PBE-OPT-Gamma': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with PBE in Gamma point \n
            '3-HSE-SCF-Gamma': Electronic SCF with HSE only in Gamma point \n
            '4-HSE-OPT-Gamma': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with HSE in Gamma point \n
            '5-PBE-SCF': Electronic SCF with PBE \n
            '6-HSE-SCF': Electronic SCF with HSE \n
            '7-HSE-OPT': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with HSE'''

        Returns:
            List of VaspJob objects
        """
        stepnames = ['1-PBE-SCF-Gamma','2-PBE-OPT-Gamma','3-HSE-SCF-Gamma',
                  '4-HSE-OPT-Gamma','5-PBE-SCF','6-HSE-SCF','7-HSE-OPT']
        jobs = []
        
        sn = 1 #set step number 1
        vaspjob = self.pbe_scf_gamma(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 2 
        vaspjob = self.pbe_ionic_rel_gamma(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 3 
        vaspjob = self.hse_scf_gamma(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 4 
        vaspjob = self.hse_ionic_rel_gamma(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)

        sn = 5 
        vaspjob = self.pbe_scf(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 6 
        vaspjob = self.hse_scf(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 7
        vaspjob = self.hse_ionic_rel(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)

        return jobs
    
    
    def hse_relaxation_gamma_extended(self,scheme_name='HSE-rel-gamma-ext'):
        """
        Generates calculation scheme for structure relaxations for HSE with more intermediate steps. Steps: \n
            '1-PBE-SCF-Gamma': Electronic SCF with PBE only in Gamma point \n
            '2-PBE-OPT-Gamma': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with PBE in Gamma point \n
            '3-HSE-SCF-Gamma': Electronic SCF with HSE only in Gamma point \n
            '4-HSE-OPT-Gamma': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with HSE in Gamma point \n
            '5-PBE-SCF': Electronic SCF with PBE \n
            '6-PBE-OPT': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with PBE \n
            '7-HSE-SCF': Electronic SCF with HSE \n
            '8-HSE-OPT': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with HSE'''

        Returns:
            List of VaspJob objects
        """
        stepnames = ['1-PBE-SCF-Gamma','2-PBE-OPT-Gamma','3-HSE-SCF-Gamma',
                  '4-HSE-OPT-Gamma','5-PBE-SCF','6-PBE-OPT','7-HSE-SCF','8-HSE-OPT']
        jobs = []
        
        sn = 1 #set step number 1
        vaspjob = self.pbe_scf_gamma(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 2
        vaspjob = self.pbe_ionic_rel_gamma(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 3
        vaspjob = self.hse_scf_gamma(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 4
        vaspjob = self.hse_ionic_rel_gamma(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)

        sn = 5 
        vaspjob = self.pbe_scf(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 6 
        vaspjob = self.pbe_ionic_rel(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 7 
        vaspjob = self.hse_scf(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 8 
        vaspjob = self.hse_ionic_rel(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)

        return jobs
                      

    def hse_relaxation_short(self,scheme_name='HSE-rel-short'):
        """
        Generates calculation scheme for ionic relaxation for HSE (short version i.e. no PBE preliminary calculation). Steps: \n
            '1-PBE-SCF': Electronic SCF with HSE \n
            '2-PBE-OPT': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with HSE

        Returns:
            List of VaspJob objects
        """
        stepnames = ['1-HSE-SCF','2-HSE-OPT']        
        jobs = []
        
        sn = 1 
        vaspjob = self.hse_scf(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 2
        vaspjob = self.hse_ionic_rel(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)
        
        return jobs

    
    def hse_vol_relaxation(self,scheme_name='HSE-Vrel'):
        """
        Generates calculation scheme for structure relaxations for HSE including cell volume relaxation. Steps: \n
            '1-PBE-SCF': Electronic SCF with PBE
            '2-PBE-OPT': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with PBE
            '3-HSE-SCF': Electronic SCF with HSE
            '4-HSE-OPT-Gamma': Relaxation of atomic positions with ISIF = 2 and EDIFFG = 0.05 eV/A with HSE
            '5-HSE-VOPT': Cell volume relaxation and ionic relaxation with HSE (ISIF=3)

        Returns:
            List of VaspJob objects
        """
        stepnames = ['1-PBE-SCF','2-PBE-OPT','3-HSE-SCF','4-HSE-OPT','5-HSE-VOPT']
        jobs = []
        
        sn = 1 
        vaspjob = self.pbe_scf(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 2
        vaspjob = self.pbe_ionic_rel(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        vaspjob.incar['LHFCALC'] = '.FALSE.'
        vaspjob.incar['ISYM'] = 2
        jobs.append(vaspjob)
        
        sn = 3 
        vaspjob = self.hse_scf(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 4 
        vaspjob = self.hse_ionic_rel(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)

        sn = 5
        vaspjob = self.hse_ionic_vol_rel(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)
        return jobs
    
    
    def hubbard_tuning(self,specie,ldauu_dict=None,u_range=(1,10),scheme_name='U_tuning'):
        """
        Generates Scheme for many calculations using different Hubbard corrections

        Parameters
        ----------
        specie : (str)
            Symbol of atomic specie to apply changes on U parameter.
        ldauu_dict : (dict), optional
            Dictionary of U parameters. Keys are pymatgen Element objects, values are the associated U parameters. The default is None.
        u_range : (Tuple), optional
            Tuple with range of U parameter to scan. The default is (1,10).
        scheme_name : (str), optional
            Name for scheme. The default is "U_tuning".

        Returns
        -------
        List of VaspJob objects
        """
        jobs = []
        
        if ldauu_dict is None:
            ldauu_dict={}
            for el in self.structure.composition.elements:
                ldauu_dict[el] = 0
        else:
            if len(ldauu_dict) != len(self.structure.composition.elements):
                raise ValueError('size of "ldauu_dict" needs to be the same as the number of species in Structure')
                
        for u in range(u_range[0],u_range[1]+1):
                        
            ldauu_dict[Element(specie)] = u
            stepname = f'U{u}'
            vaspjob = self.hubbard(ldauu_dict,add_to_job_name=stepname,add_to_job_path=stepname)
            jobs.append(vaspjob)
            
        return jobs

    
    def pbe_relaxation(self,scheme_name='PBE-rel'):
        """
        Generates calculation scheme for ionic relaxation with PBE functional. Steps: \n
            '1-PBE-SCF': Electronic SCF \n
            '2-PBE-OPT': Ionic relaxation with maximum 100 ionic steps (NSW=100), Energy change between electronic steps of 1e-05 (EDIFF=1e-05), \n
                         Force convergence criterion of 0.05 eV/Amstrong (EDIFFG=-0.05).

        Returns
        -------
        List of VaspJob objects
        """        
        stepnames = ['1-PBE-SCF','2-PBE-OPT']
        jobs = []
              
        sn = 1
        vaspjob = self.pbe_scf(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 2 
        vaspjob = self.pbe_ionic_rel(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)
        
        return jobs


    def pbe_relaxation_gamma(self,scheme_name='PBE-rel-gamma'):
        """
        Generates calculation scheme for ionic relaxation with PBE functional. Steps: \n
            '1-PBE-SCF-Gamma': Electronic SCF in Gamma point\n
            '2-PBE-OPT-Gamma': Ionic relaxation in Gamma point with maximum 100 ionic steps (NSW=100), Energy change between electronic steps of 1e-05 (EDIFF=1e-05), \n
                         Force convergence criterion of 0.05 eV/Amstrong (EDIFFG=-0.05).
            '3-PBE-SCF': Electronic SCF \n
            '4-PBE-OPT': Ionic relaxation with maximum 100 ionic steps (NSW=100), Energy change between electronic steps of 1e-05 (EDIFF=1e-05), \n
                         Force convergence criterion of 0.05 eV/Amstrong (EDIFFG=-0.05).

        Returns
        -------
        List of VaspJob objects
        """        
        stepnames = ['1-PBE-SCF-Gamma','2-PBE-OPT-Gamma','3-PBE-SCF','4-PBE-OPT']
        jobs = []

        sn = 1
        vaspjob = self.pbe_scf_gamma(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 2 
        vaspjob = self.pbe_ionic_rel_gamma(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)
              
        sn = 3
        vaspjob = self.pbe_scf(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 4 
        vaspjob = self.pbe_ionic_rel(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)
        
        return jobs
    
    
    def pbe_vol_relaxation(self,scheme_name='PBE-Vrel'):
        """
        Generates calculation scheme for ionic and cell relaxation with PBE functional. Steps: \n
            '1-PBE-SCF': Electronic SC \n
            '2-PBE-OPT': Ionic relaxation with maximum 100 ionic steps (NSW=100), Energy change between electronic steps of 1e-05 (EDIFF=1e-05), \n
                         Force convergence criterion of 0.05 eV/Amstrong (EDIFFG=-0.05) \n
            '3-PBE-VOPT': Same parameters for ionic relaxation and relaxation of the cell (ISIF=3).

        Returns
        -------
        List of VaspJob objects
        """
        stepnames = ['1-PBE-SCF','2-PBE-OPT','3-PBE-VOPT']
        jobs = []
        sn = 1
        vaspjob = self.pbe_scf(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)
        
        sn = 2 
        vaspjob = self.pbe_ionic_rel(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)
 
        sn = 3 
        vaspjob = self.pbe_ionic_vol_rel(add_to_job_name=scheme_name+'_%i'%sn ,add_to_job_path=stepnames[sn-1])
        jobs.append(vaspjob)       
        
        return jobs
                 



        
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 11 16:33:19 2023

@author: villa
"""
from pynter.testing.core import PynterTest
from pynter.testing.structure import StructureTest
from pynter.testing.hpc import JobSettingsTest
from pynter.testing.vasp import VaspInputsTest, VaspOutputsTest

class JobTest(PynterTest):
    """
    Contains methods to test Job objects
    """
    
    def __init__(self,jobclass='VaspJob'):
        """
        Specify which class the job object belongs to. Currently only "VaspJob" is available.
        """
        if jobclass == 'VaspJob':
            self.cls = VaspJobTest
        elif jobclass == 'VaspNEBJob':
            self.cls = VaspNEBJobTest
        super().__init__()
        
    def assert_inputs_equal(self,job1,job2,**kwargs):
        self.cls().assert_inputs_equal(job1,job2,**kwargs)
        return

    def assert_outputs_equal(self,job1,job2,**kwargs):
        self.cls().assert_outputs_equal(job1,job2,**kwargs)
        return
        
    def assert_settings_equal(self,job1,job2):
        JobSettingsTest().assert_job_settings_equal(job1.job_settings,job2.job_settings)
        
    def assert_job_equal(self,job1,job2,kwargs_input={},kwargs_output={}):
        self.assert_inputs_equal(job1,job2,**kwargs_input)
        if job1.outputs or job2.outputs:
            self.assert_outputs_equal(job1,job2,**kwargs_output)
        #self.assert_settings_equal(job1,job2)  # don't test for equal settings since they are system dependent
            
            
class DatasetTest(PynterTest):
    
    def assert_dataset_equal(self,ds1,ds2,kwargs_input={},kwargs_output={}):
        """
        Check if jobs in dataset are equal. Jobs are sorted before they are checked.
        """
        ds1.sort_jobs()
        ds2.sort_jobs()
        for i in range(len(ds1.jobs)):
            job1,job2 = ds1.jobs[i], ds2.jobs[i]
            JobTest(jobclass=job1.jobclass).assert_job_equal(job1, job2,kwargs_input,kwargs_output)
            
    def assert_jobs_equal(self,jobs1,jobs2,kwargs_input={},kwargs_output={}):
        """
        Check if lists of jobs are equal.
        """
        assert len(jobs1) == len(jobs2), "Job lists have different lengths"
        for i in range(0,len(jobs1)):
            j1, j2 = jobs1[i], jobs2[i]
            assert j1.jobclass == j2.jobclass, "Jobs belong to different classes" 
            JobTest(jobclass=j1.jobclass).assert_job_equal(j1, j2,kwargs_input,kwargs_output)
            
            
class VaspJobTest(JobTest):
    """
    Provides methods to test VaspJob objects
    """
    def assert_inputs_equal(self,job1,job2,include_incar=True,system_only=False):
        VaspInputsTest().assert_VaspInput_equal(job1.inputs,job2.inputs,include_incar,system_only)

    def assert_outputs_equal(self,job1,job2):
        VaspOutputsTest().assert_ComputedEntry_equal(job1.computed_entry,job2.computed_entry)
        if "Vasprun" in [job1.outputs.keys(),job2.outputs.keys()]:
            VaspOutputsTest().assert_Vasprun_equal(job1.vasprun, job2.vasprun)
            
            
class VaspNEBJobTest(JobTest):
    """
    Provides methods to test VaspNEBJob objects
    """
    def assert_inputs_equal(self,job1,job2,include_incar=True,system_only=False):
        """
        Assert incar, kpoints, potcar and images.
        """
        VaspInputsTest().assert_Incar_equal(job1.incar, job2.incar, system_only=system_only)
        VaspInputsTest().assert_Kpoints_equal(job1.kpoints, job2.kpoints)
        VaspInputsTest().assert_Potcar_equal(job1.potcar, job2.potcar)

        if len(job1.structures) != len(job2.structures):
            raise AssertionError('Number of images differ')
        else:
            for i in range(0,len(job1.structures)):
                StructureTest().assert_Structure_equal(
                    job1.structures[i], job2.structures[i])
            
    def assert_outputs_equal(self,job1,job2):
        """
        Assert pymatgen NEBAnalysis object
        """
        self.assert_object_almost_equal(job1.neb_analysis.as_dict(), job2.neb_analysis.as_dict())           
            
            
            




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 10 13:35:52 2025

@author: villa
"""
import os.path as op

from pymatgen.io.vasp.inputs import Poscar, VaspInput, Incar

from pynter.hpc.slurm import JobSettings
from pynter.jobs.vasp.vasp_jobs import VaspJob, VaspNEBJob
from pynter.vasp.schemes.core import DefaultInputs


class NEBSchemes:
    
    def __init__(self,path,structures,incar_settings=None,kpoints=None,potcar=None,job_settings=None,name=None,add_parent_folder=False):
        """
        Parameters
        ----------
        path : (str)
            Main path for the scheme.        
        structures : (list of Pymatgen Structure objects), optional
            List of structures for reaction path.
        incar_settings : (Dict), optional
            Dictionary with incar flags. The default is None. If None the default settings for PBE functional from the DefaultInputs class are used.
        kpoints : (Pymatgen Kpoints object), optional
            Pymatgen Kpoints object. The default is None. If None the default settings from the DefaultInputs class are used.
        potcar : (Pymatgen kpoints object), optional
            Pymatgen kpoints object. The default is None. If None the default settings from the DefaultInputs class are used.
        job_settings : (Dict), optional
            Dictionary with job settings to create job script, parameters are defined in ScrpitHandler class function. The default is None.\n
            If job_settings is None, the 'name' key will be added, the value is the 'name' parameter if provided, if 'name' parameter is \n
            None the value will be: 'no_name'. If automations are not desired, set job_settings['add_automation'] equal to ''.
        name : (str), optional
            Name for the system to set up scheme for. The default is None.
        add_parent_folder : (bool), optional
            Add folder to the path names like the name of the InputSets. Default is False.
        """
        
        self.path = path
        self.structures = structures
        self.incar_settings = incar_settings if incar_settings else DefaultInputs(self.structures[0]).get_incar_default()
        self.kpoints = kpoints if kpoints else DefaultInputs(self.structures[0]).get_kpoints_default()
        self.potcar = potcar if potcar else DefaultInputs(self.structures[0]).get_potcar()
        if job_settings:
            self.job_settings = JobSettings(**job_settings)
        elif name:
            self.job_settings = JobSettings(slurm={'job-name':name})
        else:
            self.job_settings = JobSettings(slurm={'job-name':'no_name'})
        self.name = name if name != None else self.job_settings['job-name'] 
        
        if self.name: # this return false if name is '', the previuos line considers only if name is None
            self.job_settings['job-name'] = self.name

        if add_parent_folder:
            self.path = op.join(self.path,self.name)

        for key in ['KPAR','NCORE','NPAR']:
            if key in self.incar_settings:
                del self.incar_settings[key]

        self.images = len(self.structures)-2


    def cineb_pbe(self,nionic_steps=50,scheme_name=None,stepnames=['CINEB']):
        """
        Climbing image NEB job with PBE. The force convergence is set to 0.05 eV/A, relaxation is done with damped 
        dynamics (IBRION=3), symmetry is turned off (ISYM=0). The maximum number of ionic steps is set to nionic_steps.
        """
        if not scheme_name:
            scheme_name = stepnames[0]

        nebjob = self.neb_generic_pbe(scheme_name=scheme_name,stepnames=stepnames)
        nebjob.incar.update({
            'ISYM': 0,
            'EDIFF': 1e-05,
            'EDIFFG': -0.05,
            'ISIF': 2,
            'NSW' : nionic_steps,
            'IOPT' : 0,
            'IBRION' : 3,
            'POTIM' : 0.05,
            'LCLIMB':'.TRUE.'
            })

        return nebjob

 
    def neb_complete_pbe(self,nionic_steps=50,scheme_name=None):
        """
        NEB scheme with 3 steps:
            - Step 1: preconvergence (SCF) of the images with EDIFF = 1e-04
            - Step 2: NEB calculation with NSW = nionic_steps
            - Step 3: CINEB with NSW = nionic_steps
        """
        scheme_name = scheme_name if scheme_name != None else 'NEB_comp'
        jobs = []
        step1 = self.preconverge(scheme_name='NEB-SCF_1',stepnames=['1-SCF'])
        for j in step1:
            jobs.append(j)
        
        step2 = self.neb_pbe(nionic_steps=nionic_steps,scheme_name='NEB_2',stepnames=['2-NEB'])
        jobs.append(step2)
        
        step3 = self.cineb_pbe(nionic_steps=nionic_steps,scheme_name='NEB_3',stepnames=['3-CINEB'])
        jobs.append(step3)

        return jobs
 

    def neb_complete_4step_pbe(self,nionic_steps=50,scheme_name=None):
        """
        NEB scheme with 4 steps:
            - Step 1: preconvergence (SCF) of the images with EDIFF = 1e-04
            - Step 2: NEB calculation with NSW = 10
            - Step 3: NEB calculation with NSW = nionic_steps
            - Step 4: CINEB with NSW = nionic_steps
        """
        scheme_name = scheme_name if scheme_name != None else 'NEB_comp'
        jobs = []
        step1 = self.preconverge(scheme_name='NEB-SCF_1',stepnames=['1-SCF'])
        for j in step1:
            jobs.append(j)
        
        step2 = self.neb_pbe(scheme_name='NEB_2',stepnames=['2-NEB-NSW10'])
        step2.inputs['INCAR']['NSW'] = 10
        jobs.append(step2)

        step3 = self.neb_pbe(nionic_steps=nionic_steps,scheme_name='NEB_3',stepnames=['3-NEB'])
        jobs.append(step3)
        
        step4 = self.cineb_pbe(nionic_steps=nionic_steps,scheme_name='NEB_4',stepnames=['4-CINEB'])
        jobs.append(step4)

        return jobs


    def neb_generic_pbe(self,scheme_name=None,stepnames=['NEB-generic']):
        """
        Returns a VaspNEBJon with just necessary changes to INCAR. No particular calculation scheme is set.
        """
        if not scheme_name:
            scheme_name = stepnames[0]
        inputs = {}
        incar_settings = self.incar_settings.copy()
        job_settings = self.job_settings.copy()

        incar_settings.update({
            'IMAGES' : self.images,
            'SPRING' : -5,
            'IOPT' : 0, #default
            'ICHAIN' : 0, #default 
            'LCLIMB' : '.FALSE.', 
            'LTANGENTOLD' : '.FALSE.', #default
            'LDNEB' : '.FALSE.', #default
            'LNEBCELL' : '.FALSE.' #default
            })

        incar = Incar(incar_settings)
        kpoints = self.kpoints
        potcar = self.potcar
        
        inputs['structures'] = self.structures
        inputs['INCAR'] = incar
        inputs['KPOINTS'] = kpoints
        inputs['POTCAR'] = potcar        
        job_settings['job-name'] = '_'.join([self.name,scheme_name])

        jobname = '_'.join([self.name,scheme_name])
        jobpath = op.join(self.path,stepnames[0])
        nebjob = VaspNEBJob(path=jobpath,inputs=inputs,job_settings=job_settings,name=jobname)

        return nebjob       
    

    def neb_pbe(self,nionic_steps=50,scheme_name=None,stepnames=['NEB']):
        """
        Standard NEB job with PBE. The force convergence is set to 0.1 eV/A, relaxation in done with damped 
        dynamics (IBRION=3), symmetry is turned off (ISYM=0). The maximum number of ionic steps is set to 200.
        """
        if not scheme_name:
            scheme_name = stepnames[0]

        nebjob = self.neb_generic_pbe(scheme_name=scheme_name,stepnames=stepnames)
        nebjob.incar.update({
            'ISYM': 0,
            'EDIFF': 1e-04,
            'EDIFFG': -0.1,
            'ISIF': 2,
            'NSW' : nionic_steps,
            'IBRION' : 3,
            'POTIM' : 0.05,
            })

        return nebjob


    def preconverge(self,scheme_name=None,stepnames=['NEB-preconverge']):
        """
        Do SCF convergence for all images before starting NEB calculation
        """      
        scheme_name = scheme_name if scheme_name!=None else 'SCF'

        self.incar_settings.update({
            'EDIFF' : 1e-04,
            'NSW': 0
            })

        jobs = []
        structures = self.structures
        for s in structures:
            
            index = structures.index(s)
            image_name = str(index).zfill(2)
            
            incar = Incar(self.incar_settings)
            kpoints = self.kpoints
            potcar = self.potcar
            poscar = Poscar(s)
            vaspinput = VaspInput(incar, kpoints, poscar, potcar)
            
            job_settings = self.job_settings.copy()
            job_settings['job-name'] = '_'.join([self.name,scheme_name,image_name])
            
            jobname = '_'.join([self.name,scheme_name,image_name])
            jobpath = op.join(self.path,stepnames[0],image_name)
            vaspjob = VaspJob(path=jobpath,inputs=vaspinput,job_settings=job_settings,name=jobname)
            jobs.append(vaspjob)
            
        return jobs      
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Sep  8 19:20:55 2023

@author: lorenzo
"""
import numpy as np
import warnings

from pymatgen.io.vasp.inputs import Potcar


def get_charge_from_computed_entry(entry):
    """
    Get charge of vasp calculation from ComputedEntry object. 
    Subtracts the total valence electrons from potcar to 'NELECT' value
    """
    potcar_symbols_bulk = [d['titel'].split(' ')[1] for d in entry.parameters['potcar_spec']]
    potcar = Potcar(potcar_symbols_bulk)
    charge = 0
    if 'parameters' in entry.data.keys():
        parameters = entry.data['parameters']
    else:
        raise ValueError('"parameters" need to be present in ComputedEntry data')
    if 'NELECT' in parameters.keys():
        nelect = parameters['NELECT']
        val = {}
        for p in potcar:
            val[p.element] = p.nelectrons
        neutral = sum([ val[el.symbol]*coeff 
                       for el,coeff in entry.structure.composition.items()])
        charge = neutral - nelect
    if not isinstance(charge,int):
        charge = np.around(charge,decimals=1)
        
    return charge


def get_convergence_from_outcar(file='OUTCAR'):
    """
    Check convergence of VASP calculation from OUTCAR. The following logic is used:
    1. Check for lines in OUTCAR: 
        - Every line with "reached required accuracy" is a succesful electronic step
        - "stopping structural energy minimisation" indicates completed ionic convergence
        - "General timing and accounting information for this job" indicates that the job
          has gone to completion.
    2. Convergence criteria:
        - Electronic convergence is achieved if number_electronic_steps < 'NELM' parameter
        - if 'IBRION' is -1 (no ionic relaxation) also ionic convergence returns True
        - if 'IBRION' is one of (0,1,2,3) ionic convergence is returned based on OUTCAR line
        - If 'IBRION' > 3 ionic convergence returns None
        
    Parameters
    ----------
    file : (str)
        Path to OUTCAR file.

    Returns
    -------
    is_converged_electronic (bool)
    is_converged_ionic (bool)
    """
    try:
        with open(file, "r") as f:
            lines = f.readlines()      
        parameters = {}
        is_converged_electronic, is_converged_ionic, is_job_finished = False,False,False
        n_electronic_steps = 0
        for line in lines:    
            if "reached required accuracy" in line:
                n_electronic_steps += 1
            if "stopping structural energy minimisation" in line:
                is_converged_ionic = True
            if "General timing and accounting informations for this job" in line:
                is_job_finished = True
            param = _get_parameter_from_outcar_line(tags=['IBRION','NELM'],outcar_line=line)
            if param:
                parameters.update(param)
        
        ibrion = int(parameters['IBRION'])
        nelm = int(parameters['NELM'])
        if is_job_finished:            
            if n_electronic_steps < nelm:
                is_converged_electronic = True
            if ibrion in [0,1,2,3]:
                return  is_converged_electronic, is_converged_ionic
            elif ibrion == -1 :
                return is_converged_electronic, True
            else:
                warnings.warn('Fast convergence check from OUTCAR not yet implemented for IBRION > 3')
                return is_converged_electronic, None
        else:
            return is_converged_electronic, is_converged_ionic

    except FileNotFoundError:
        warnings.warn(f'{file} not found')
        return None,None


def get_parameters_from_outcar(tags,file='OUTCAR'):    
    """
    Get parameters from OUTCAR file based on target line search
    """
    parameters = {}
    try:
        with open(file,'r') as f:
            lines = f.readlines()
        for line in lines:
            param = _get_parameter_from_outcar_line(tags=tags,outcar_line=line)
            if param:
               parameters.update(param)
        return parameters
    except FileNotFoundError:
        return None
    
                                
def _get_parameter_from_outcar_line(tags,outcar_line):
    """
    Get parameter value from a line in OUTCAR file based on target line search.
    Lines are defined in the parameter_strings dict.
    """
    paramater_strings = {
        'IBRION': 'ionic relax: 0-MD 1-quasi-New 2-CG',
        'NSW':'number of steps for IOM',
        'NELM':'# of ELM steps'
        }
    params = {}
    for key in tags:
        target = paramater_strings[key]
        if target in outcar_line:
            characters_without_spaces = [char for char in outcar_line.split(' ') if char != '']
            tag = characters_without_spaces[0]
            if key != tag:
                raise ValueError('Target tag and tag read from OUTCAR differ')
            value = characters_without_spaces[2].replace(';','')
            params[tag] = value
    return params

    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Mar 29 17:58:14 2025

@author: lorenzo
"""
import os.path as op
from shutil import copyfile


def are_kpoints_equal(job1,job2):
    """
    Compare KPOINTS btw 2 VaspJob objects.
    """               
    path1 = op.abspath(job1.path)
    path2 = op.abspath(job2.path)
    kpoints1 = job1.kpoints.as_dict() 
    kpoints2 = job2.kpoints.as_dict()
    relative_path1 = op.relpath(path1)
    relative_path2 = op.relpath(path2)
    if kpoints1 == kpoints2:
        message = f'KPOINTS in {relative_path1} and {relative_path2} are the same'
        are_kpts_equal = True
    else:
        message = f'KPOINTS in {relative_path1} and {relative_path2} are different'
        are_kpts_equal = False
    return are_kpts_equal, message

def copy_file_from_jobs(filename,job1,job2):
    """
    Copy file from path of Job1 to path of Job2.
    """
    source = op.join(op.abspath(job1.path),filename)
    destination = op.join(op.abspath(job2.path),filename)
    copyfile(source,destination)
    return f'{op.relpath(source)} copied to {op.relpath(destination)}'
    
def copy_contcar_to_poscar(job1,job2):
    """
    Copy CONTCAR of Job1 to POSCAR of Job2
    """
    path1 = op.abspath(job1.path)
    path2 = op.abspath(job2.path)
    source = op.join(path1,'CONTCAR')
    destination = op.join(path2, 'POSCAR')
    copyfile(source, destination)
    return f'{op.relpath(source)} copied to {op.relpath(destination)}'
    
def is_limit_electronic_steps_reached(job):
    """
    Check if electronic step limit has been reached
    """
    if job.is_converged_electronic is not None:
        ionic_steps = job.vasprun.ionic_steps
        n_electronic_steps = len(ionic_steps[-1]['electronic_steps'])
        nelm = job.vasprun.parameters['NELM']        
        return True if n_electronic_steps == nelm else False   
    else:
        return False

def is_limit_ionic_steps_reached(job):
    """
    Check if ionic step limit has been reached
    """
    if job.is_converged_ionic is not None:
        n_ionic_steps = len(job.vasprun.ionic_steps)
        nsw = job.vasprun.parameters['NSW']
        return True if nsw == n_ionic_steps else False 
    else:
        return False
    
    
    
class VaspAutomation:
    
    def __init__(self,status_filename='exit_status.txt'):
        """
        Class to handle automation workflows for VASP

        Parameters
        ----------
        status_filename: (str)
            Filename to write automation output. The default is 'exit_status.txt'.
        """
        self.status = []
        self.status_filename = status_filename
    
    
    def check_convergence(self,job):
        """
        Check electronic and ionic convergence and update status
        """
        conv_el, conv_ionic = job.is_converged_electronic, job.is_converged_ionic
        self.status.append('Analysing "vasprun.xml"... ')
        self.status.append(f'\nElectronic convergence: {conv_el}')
        self.status.append(f'Ionic convergence: {conv_ionic}\n') 
        if conv_el == False and conv_ionic == False:
            self.status.append('Calculation probably did not converge, control manually')
        return conv_el,conv_ionic
                        
    
    def check_step_limits_reached(self,job):
        """
        Check if electronic or ionic steps limit has been reached. \n
        Electronic steps limit is taken from the "NELM" flag in the INCAR file. \n
        Ionic steps limit is taken from the "NSW" flag in the INCAR file

        Returns
        -------
        el_limit_reached : (bool)
        
        ionic_limit_reached : (bool)
        """
        limit_el_steps_reached, limit_ionic_steps_reached = False, False
        if is_limit_electronic_steps_reached:
            limit_el_steps_reached = True
            self.status.append('Maximum limit of electronic steps has been reached')
        if self.limit_ionic_steps_reached():
            limit_ionic_steps_reached = True
            self.status.append('Maximum limit of ionic steps has been reached')         
        return limit_el_steps_reached, limit_ionic_steps_reached        
 
    
    def copy_files_from_job1_to_job2(self,
                                     job1,
                                     job2,
                                     filenames,
                                     check_kpoints=False):
        """
        Copy VASP files from directory of Job1 to directory of Job2

        Parameters
        ----------
        job1 : (VaspJob)
            VaspJob object.
        job2 : (VaspJob)
            VaspJob object.
        filenames : (list)
            List of VASP file names to copy. If "CONTCAR" is present, it will be copied 
            as "POSCAR" for Job2.
        check_kpoints : (bool)
            Only copy CHGCAR and WAVECAR if the KPOINTS file are the same btw the two jobs.
        """
        kpoints_equal = True
        if check_kpoints:
            kpoints_equal, kpoints_message = are_kpoints_equal(job1, job2)
            self.status.append(kpoints_message)
        for file in filenames:
            if file=='CONTCAR':
                message = copy_contcar_to_poscar(job1, job2)
                self.status.append(message)
                
            elif file in ['CHGCAR','WAVECAR']:    
                if kpoints_equal or check_kpoints is False:
                    message = copy_file_from_jobs(file, job1, job2)
                    self.status.append(message)
                else:
                    self.status.append(f'Copying of {file} ignored because KPOINTS do not match ')
            
            else:
                message = copy_file_from_jobs(file, job1, job2)
                self.status.append(message)   
    
        
    def write_status(self,path,filename=None):
        """
        Write status lines to file

        Parameters
        ----------
        path : (str), optional
            Path of file to write.
        filename : (str), optional
            Filename. The default is None. If None "status_filename" attribute is used.
        """

        filename = filename or self.status_filename
        if filename:
            with open(filename,'w') as f:
                f.write('\n'.join(self.status))
                print('Automation report written in "%s" file in directory "%s"' %(filename,path))
        else:
            print('Writing status file is disabled, if needed please set "status_filename" kwarg or "--status" optional argument on command line')
            
            
            
class VaspNEBAutomation(VaspAutomation):
    
    def copy_images_files_from_job1_to_job2(self,job1,job2,filenames=['CHGCAR','CONTCAR','WAVECAR']):
        """
        Copy VASP files from directory of Job1 to directory of Job2

        Parameters
        ----------
        job1 : (VaspJob)
            VaspJob object.
        job2 : (VaspJob)
            VaspJob object.
        filenames : (list)
            List of VASP file names to copy. If "CONTCAR" is present, it will be copied 
            as "POSCAR" for Job2.
        """
        path2 = job2.path
        for image_path in job1.image_dirs:
            image_name = op.path.basename(image_path)
            for file in filenames:
                if file == 'CONTCAR':
                    source = op.join(image_path,'CONTCAR')
                    dest = op.join(path2,image_name,'POSCAR')    
                    copyfile(source,dest)
                else:
                    source = op.join(image_path,file)
                    dest = op.join(path2,image_name,file)
                    copyfile(source,dest)                
                self.status.append(f'{op.relpath(source)} copied in {op.relpath(dest)}')
        
        
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 10 13:57:21 2025

@author: villa
"""
import os.path as op

from defermi.generator import create_vacancies, create_substitutions
from defermi.structure import defect_finder
from pynter.vasp.schemes.core import InputSets
from pynter.vasp.schemes.relaxation import RelaxationSchemes
from pynter.vasp.schemes.variations import VariationsSchemes


class DefectSchemes(InputSets):
            
    
    def defects_pbe_relaxation(self,defects_with_charges,locpot=False,rel_scheme='default'):
        """
        Generate jobs for default defect calculation scheme with PBE.

        Parameters
        ----------
        defects_with_charges : (list)
            List of tuples with Defect object and relative charge state list, for example [(Defect , [q1,q2,q3,...])].
        locpot : (bool), optional
            Add 'LVTOT=True' in INCAR. The default is False.
        rel_scheme : (str), optional
            Relaxation scheme to use with PBE. 'default for pbe_relaxation and 
            'gamma' pbe_relaxation_gamma. The default is 'default'.

        Returns
        -------
        jobs : (list)
            List of VaspJob objects.
        """
        jobs = []

        for df, charge_states in defects_with_charges:
            defect_name = df.name.replace('(','_').replace(')', '') #replace label parenthesis
            path = op.join(self.path,defect_name)
            schemes_q = VariationsSchemes(path=path,structure=df.defect_structure,incar_settings=self.incar_settings,
                                job_settings=self.job_settings,name=self.name+'_'+defect_name,add_parent_folder=False)
            charge_jobs = schemes_q.charge_states(charge_states,locpot)
            for jq in charge_jobs:
                schemes_rel = RelaxationSchemes(path=jq.path,vaspinput=jq.inputs,job_settings=jq.job_settings,name=jq.name)
                if rel_scheme == 'default':
                    rel_jobs = schemes_rel.pbe_relaxation()
                elif rel_scheme == 'short':
                    rel_jobs = schemes_rel.pbe_ionic_rel(add_to_job_name=None,add_to_job_path=None)
                elif rel_scheme == 'gamma':
                    rel_jobs = schemes_rel.pbe_relaxation_gamma()
                for jrel in rel_jobs:
                        jobs.append(jrel)
        return jobs


    def defects_pbe_relaxation_from_structures(self,structures_with_charges,tol=1e-03,locpot=False,rel_scheme='default'):
        """
        Generate jobs for default defect calculation scheme with PBE starting from defect structures.
        defect_finder is used to create defect objects, check that the algorithm wirks as intended.

        Parameters
        ----------
        structures_with_charges : (list)
            List of tuples with Structure object and relative charge state list, for example [(Structure , [q1,q2,q3,...])].
        tol: (float)
            Tolerance for defect_finder.
        locpot : (bool)
            Add 'LVTOT=True' in INCAR. The default is False.
        rel_scheme : (str)
            Relaxation scheme to use with PBE. 'default' for pbe_relaxation and 
            'gamma' pbe_relaxation_gamma. The default is 'default'.

        Returns
        -------
        jobs : (list)
            List of VaspJob objects.
        """
        defects_with_charges = []
        for structure_defect, charge_states in structures_with_charges:
            defect = defect_finder(structure_defect, self.structure, tol=tol)
            defects_with_charges.append( (defect,charge_states) )
        return self.defects_pbe_relaxation(defects_with_charges,locpot=locpot,rel_scheme=rel_scheme)


    def interstitials_pbe_relaxation(self):
        """
        Since the interstitials structures neeed to be checked and selected, calculations should be set up manually 
        with defects_pbe_relaxation. If the function needs to be implemented, set up a criterion for the choice of 
        interstitials sites.
        """
        pass

            
    def substitutions_pbe_relaxation(self,elements_to_replace_with_charges,supercell_size=None,locpot=False,rel_scheme='default'):
        """
        Generate jobs for default substitutions calculation scheme with PBE.

        Parameters
        ----------
        elements_to_replace_with_charges : (dict)
            Dictionary which indicates the substitution types and the relative charge states list.
            Needs to be in this format:
                {'<new_el>-on-<old_el>':[q0,q1,q2]}, where <new_el> and <old_el> are the symbols of the element to be added and
                the element to be substituted with, respectively. [q0,q1,q2] is the list of corrisponding charge states for this substitution.
            The format of the substitution string is the same that is generated from the function "create_substitutiion_structures".
        supercell_size : (int or numpy array), optional
            Input for the make_supercell function of the Structure class. The default is 1 (the structure is not modified).
        locpot : (bool), optional
            Add 'LVTOT=True' in INCAR. The default is False.
        rel_scheme : (str), optional
            Relaxation scheme to use with PBE. 'default for pbe_relaxation and 
            'gamma' pbe_relaxation_gamma. The default is 'default'.

        Returns
        -------
        jobs : (list)
            List of VaspJob objects.
        """
        defects_with_charges = []
        for key,charge_states in elements_to_replace_with_charges.items():
            elements_to_replace = {key.split('-on-')[1] : key.split('-on-')[0]}
            substitution = create_substitutions(self.structure,elements_to_replace=elements_to_replace,
                                                supercell_size=supercell_size)[0]
            defects_with_charges.append((substitution,charge_states))
        jobs = self.defects_pbe_relaxation(defects_with_charges,locpot=locpot,rel_scheme=rel_scheme)
        return jobs        
        

    def vacancies_pbe_relaxation(self,elements_with_charges,supercell_size=None,locpot=False,rel_scheme='default'):
        """
        Generate jobs for default vacancies calculation scheme with PBE.

        Parameters
        ----------
        elements_with_charges : (dict)
            Dictionary with element symbols as keys and relative charge states (int) lists as values ({'el':[-1,0,1]}).
        supercell_size : (int or numpy array), optional
            Input for the make_supercell function of the Structure class.
            If None the input structure is not modified. The default is None.
        locpot : (bool), optional
            Add 'LVTOT=True' in INCAR. The default is False.
        rel_scheme : (str), optional
            Relaxation scheme to use with PBE. 'default for pbe_relaxation and 
            'gamma' pbe_relaxation_gamma. The default is 'default'.

        Returns
        -------
        jobs : (list)
            List of VaspJob objects.
        """
        defects_with_charges = []
        for el,charge_states in elements_with_charges.items():
            vacancy = create_vacancies(structure=self.structure,elements=[el],supercell_size=supercell_size)[0]
            defects_with_charges.append((vacancy,charge_states))
        jobs = self.defects_pbe_relaxation(defects_with_charges,locpot=locpot,rel_scheme=rel_scheme)
        return jobs


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 27 10:52:08 2020

@author: villa
"""

import numpy as np
import os.path as op
import os
import collections
from scipy.spatial import KDTree
from ase.visualize import view
from pymatgen.util.coord import pbc_shortest_vectors
from pymatgen.core.periodic_table import Element
from pymatgen.core.composition import Composition
from pymatgen.core.structure import Structure
from pymatgen.core.trajectory import Trajectory

def _get_distance_vector_and_image(lattice,frac_coords1,frac_coords2,jimage=None):
    """
    Same as pymatgen.core.Lattice.get_distance_and_image but returns 
    the distance vetor instead of the norm.
    Gets distance between two frac_coords assuming periodic boundary
    conditions. If the index jimage is not specified it selects the j
    image nearest to the i atom and returns the distance and jimage
    indices in terms of lattice vector translations. If the index jimage
    is specified it returns the distance between the frac_coords1 and
    the specified jimage of frac_coords2, and the given jimage is also
    returned.
    Args:
        frac_coords1 (3x1 array): Reference fcoords to get distance from.
        frac_coords2 (3x1 array): fcoords to get distance from.
        jimage (3x1 array): Specific periodic image in terms of
            lattice translations, e.g., [1,0,0] implies to take periodic
            image that is one a-lattice vector away. If jimage is None,
            the image that is nearest to the site is found.
    Returns:
        (distance, jimage): distance and periodic lattice translations
        of the other site for which the distance applies. This means that
        the distance between frac_coords1 and (jimage + frac_coords2) is
        equal to distance.
    """
    if jimage is None:
        v, d2 = pbc_shortest_vectors(lattice, frac_coords1, frac_coords2, return_d2=True)
        fc = lattice.get_fractional_coords(v[0][0]) + frac_coords1 - frac_coords2  # type: ignore
        fc = np.array(np.round(fc), dtype=int)
        return v, fc

    jimage = np.array(jimage)
    mapped_vec = lattice.get_cartesian_coords(jimage + frac_coords2 - frac_coords1)  # type: ignore
    return mapped_vec, jimage  # type: ignore


def deform_lattice(structure,stdev=0.03):
    """
    Deform lattice vectors in structure. Also changes the angles of lattice vectors (off diagonal components).
    The function "apply_strain" from the Structure class only changes the base vectors' size but not the angles.

    Parameters
    ----------
    structure : 
        Structure object.
    stdev : float
        Standard deviation.

    Returns
    -------
    (Structure)
        Structure with deformed lattice.
    """
    def get_random_strain(stdev):
        D = np.eye(3) #strain matrix
        for i in range(3):
            D[i, i] += np.random.uniform(-stdev, stdev)
            for j in range(i + 1, 3):
                D[i, j] += np.random.uniform(-stdev, stdev) / 4
                D[j, i] = D[i, j]
        return D
    strain_matrix = get_random_strain(stdev=stdev)
    atoms = structure.to_ase_atoms()
    new_cell = atoms.get_cell().dot(strain_matrix)
    atoms.set_cell(new_cell,scale_atoms=True)
    return Structure.from_ase_atoms(atoms)


def get_distance_vector(site1,site2,jimage=None):
    """
    Get distance vector between two sites assuming periodic boundary conditions.
    Same as pymatgen.core.sites.Site.distance but returns a vector instead of norm.
    Args:
        other (PeriodicSite): Other site to get distance from.
        jimage (3x1 array): Specific periodic image in terms of lattice
            translations, e.g., [1,0,0] implies to take periodic image
            that is one a-lattice vector away. If jimage is None,
            the image that is nearest to the site is found.
    Returns:
        distance (float): Distance between the two sites
    """
    return _get_distance_vector_and_image(site1.lattice, site1.frac_coords, site2.frac_coords,jimage=jimage)[0]


def get_displacement_vectors(structures):
    """
    Get vectors in cartesian coords of site displacements w.r.t. the first structure.

    Parameters
    ----------
    structures : 
        list of Pymatgen Structure objects.

    Returns
    -------
    (numpy ndarray)
        Displacement vectors in cartesian coordinates.
    """
    traj = Trajectory.from_structures(structures,constant_lattice=True) #order of structures determines ref in traj
    traj.to_displacements()
    disp = traj.coords[1]
    return structures[0].lattice.get_cartesian_coords(disp)


def get_furthest_neighbors(structure, target_site, species):
    """
    Get the indexes of atomic sites that are furthest away from the target site
    and maximally spaced apart from each other, considering periodic boundary conditions.

    Parameters:
    - structure: pymatgen Structure object
    - target_site: a pymatgen PeriodicSite or index of the target site in the structure
    - species: list of species to find (ordered as given)

    Returns:
    - List of indices corresponding to the furthest sites of the given species
    """
    if isinstance(target_site, int):
        target_site = structure[target_site]

    selected_indices = []
    
    for specie in species:
        max_score = -1
        best_index = None

        for i, site in enumerate(structure):
            if site.species_string != specie or i in selected_indices:
                continue
            
            # Distance from target
            distance_to_target = structure.get_distance(structure.index(target_site),i)

            # Compute minimum distance to already selected sites
            if selected_indices:
                distances_to_selected = [structure.get_distance(j,i) for j in selected_indices]
                min_distance_to_selected = min(distances_to_selected)
            else:
                min_distance_to_selected = distance_to_target

            # Define score: prioritize sites far from target and far from other selections
            score = distance_to_target + min_distance_to_selected

            if score > max_score:
                max_score = score
                best_index = i
        
        if best_index is not None:
            selected_indices.append(best_index)

    return selected_indices

def get_WarrenCowley_order_parameter(
                        structure,
                        A_symbols,
                        X_symbols,
                        neighbors_cutoff):
    """
    Compute Warren Cowley order parameter to describe mixing tendencies in alloys

    The WC order parameter is computed as:
                alpha_{A-X} = 1 - < P_{A-X} > / C_X
    where:
    - <P_{A-X}> is the avg probability of finding X near A, computed as:
        <P_{A-X}> = n_X / n_total , n = number of nighbors of A
        C_X = concentration of X in structure
    
    Parameters
    ----------
    structure:
        Pymatgen Structure
    A_symbols: (list)
        List of element symbols to consider for A-site
    X_symbols: (list)
        List of element symbols to consider for X-site
    neighbors_cutoff: (float)
        Distance cutoff for neighbors counting

    Returns
    -------
    alpha: (float)
        Warren-Cowley order parameter
    """
    total_P_AX = 0
    n_A_sites = 0
    n_X_sites = 0
    for site in structure:
        if site.specie.symbol in A_symbols:
            n_A_sites += 1
            neighbors_list = structure.get_neighbors(site,r=neighbors_cutoff)
            n_X_neighbors = 0
            n_total_neighbors = 0
            for neighbor in neighbors_list:
                if neighbor.specie.symbol in X_symbols:
                    n_X_neighbors += 1
                    n_total_neighbors += 1
                elif neighbor.specie.symbol in A_symbols:
                    n_total_neighbors += 1
            total_P_AX += n_X_neighbors / n_total_neighbors
        elif site.specie.symbol in X_symbols:
            n_X_sites += 1

    if n_A_sites == 0 or n_X_sites == 0:
        raise ValueError("No A or X sites found to compute order parameter.")

    C_X = n_X_sites / (n_A_sites + n_X_sites)
    avg_P_AX = total_P_AX / n_A_sites
    alpha = 1 - avg_P_AX / C_X 
    return alpha


def is_site_in_structure(site,structure,tol=1e-03):
    """
    Check if Site is part of the Structure list. This function is needed because 
    sometimes doing a simple check ("site in structure") doesn't work. This function performes
    a check on the coordinates and the element on the site. Therefore it is more reliable.

    Parameters
    ----------
    site : (Site)
        PeriodicSite or Site object.
    structure : (Structure)
        Pymatgen Structure object.
    tol : (float), optional
        Tolerance for fractional coordinates. The default is 1e-03.

    Returns
    -------
    is_site_in_structure : (bool)
    index : (int)
        Index of site in structure in case site is_site_in_structure returns True
        If False index will be None.
    """
    is_site_in_structure,index = False,None
    check,index = is_site_in_structure_coords(site, structure,tol=tol)
    if check:
        s = structure[index]
        if site.specie.symbol == s.specie.symbol:
            is_site_in_structure =True
    
    return is_site_in_structure,index


def is_site_in_structure_coords(site, structure, tol=1e-3, return_distance=False):
    """
    Check if a site's coordinates are present in a structure, using periodic boundary conditions.

    Parameters
    ----------
    site : (Site)
        PeriodicSite or Site object.
    structure : (Structure)
        Pymatgen Structure object.
    tol : (float), optional
        Tolerance for site comparison, normalized with respect to lattice size. 
        Default is 1e-03.
    return_distance : (bool)
        Return distance btw site coords and closest site in reference structure 

    Returns
    -------
    is_site_in_structure_coords : (bool)
        Whether the site exists in the structure within tolerance.
    index : (int or None)
        Index of matching site in structure if found, otherwise None.
    distance : (float)
        Returned if return_distance is set to True. 
        distance btw site coords and closest site in reference structure.
    """
    # Normalize tolerance with lattice vector size
    l = site.lattice
    tol = np.sqrt(l.a**2 + l.b**2 + l.c**2) * tol

    # Convert structure sites to fractional coordinates
    frac_coords_structure = np.array([s.frac_coords % 1 for s in structure]) # ensure no negative coords
    kdtree_structure = KDTree(frac_coords_structure, boxsize=1.0)  # Take periodicity into account

    # Query the KDTree for nearest neighbor
    dist, index = kdtree_structure.query(site.frac_coords, distance_upper_bound=tol)

    if return_distance:
        is_site = [None,None,dist]
    else:
        is_site = [None,None]

    if dist < tol:
        is_site[0], is_site[1] = True, index
    else:
        is_site[0], is_site[1] = False, None
    
    return is_site


def rattle_atoms(structure,stdev=0.05,seed=None):
    """
    Change randomly atomic positions. Uses the ASE.
    Similar to the method "perturb" from the Structure class
    Parameters
    ----------
    structure :
        Structure object
    stdev : float
        Standard deviation
    seed : int
        Seed for random number generation passed to the Atoms.rattle method.
    """
    atoms = structure.to_ase_atoms()
    atoms.rattle(stdev=stdev,seed=seed)
    return Structure.from_ase_atoms(atoms)


def remove_oxidation_state_from_site(site):
    """
    Remove oxidation state decoration from site.
    """
    new_sp = collections.defaultdict(float)
    for el, occu in site.species.items():
        sym = el.symbol
        new_sp[Element(sym)] += occu
    site.species = Composition(new_sp)
    return

def sort_sites_to_ref_coords(structure,structure_ref,extra_sites=[],tol=1e-03,get_indexes=False):
    """
    Sort Sites of one structure to match the order of coordinates in a reference structure. 

    Parameters
    ----------
    structure : (Structure)
        Structure to sort.
    structure_ref : (Structure)
        Reference Structure.
    extra_sites : (list), optional
        Sites to append at the end of the structure. The default is [].
    tol: (float)
        Tolerance for site comparison. The distance between sites in target and reference stucture is used, 
        periodicity is accounted for. The tolerance is normalized with respect to lattice vector size. 
        The default is 1e-03.
    get_indexes : (bool), optional
        Get list of mapping indexes for target structure sites in reference structure. The default is False.

    Returns
    -------
    new_structure : (Structure)
        Sorted Structure
    indexes : (list)
        If get_indexes is True a tuple is returned. List of mapping indexes for target structure sites
        in reference structure.
    """
    df = structure
    bk = structure_ref
    indexes = []
    new_sites=[]
    for s in df:
        check,index = is_site_in_structure_coords(s,bk,tol=tol)
        if check:
            indexes.append(index)
    for w in range(0,len(bk)):
        new_sites.insert(indexes[w],df[w])
    for s in extra_sites:
        new_sites.append(s)
            
    new_structure = df.copy()
    new_structure._sites = new_sites 
    if get_indexes:
        return new_structure, indexes
    else:
        return new_structure


def view_structures_with_ase(structures):
    """
    Visualize a Structure object or a list of Structure objects with the ASE. 
    First the Structure objects are converted into an ase Atom object, then "view" is used to visualize them.
    """
    if type(structures) == list:
        atoms=[]
        for s in structures:
            atoms.append(s.to_ase_atoms())
    else:
        atoms = structures.to_ase_atoms()
    view(atoms)
    return


def write_extxyz_file(file,structure,structure_ref=None,displacements=False):
    """
    Write extxyz format. Displacements can be included for visualization in OVITO. 

    Parameters
    ----------
    file : (str)
        Path to save extxyz file to.
    structure : (Structure)
        Structure to visualize.
    structure_ref : (Structure)
        Reference Structure if needed. The default is None.
    displacements : (bool)
        Include displacement vectors. The default is False.
    """
    atoms = structure.to_ase_atoms()
    if displacements:
        disp = get_displacement_vectors([structure_ref, structure])
        atoms.arrays.update({'disp':disp})
    if not op.exists(op.dirname(file)):
        os.makedirs(op.dirname(file))
    atoms.write(file,format='extxyz')
    return


def write_xdatcar_from_structures(structures,file='XDATCAR'):
    """
    Write XDATCAR file from a list of structures. The first structure determines the reference for the Trajectory object.
    """
    traj = Trajectory.from_structures(structures,constant_lattice=True) 
    if not op.exists(op.dirname(file)):
        os.makedirs(op.dirname(file))
    traj.write_Xdatcar(file)
    return
    
    
    
    
    
    
    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May  8 17:48:31 2025

@author: villa
"""

from pynter.lammps.inputs import LammpsInput
from pynter.lammps.schemes.core import InputSets


class PACEActiveLearningSchemes(InputSets):
    
    
    def cook_and_quench(self,
                        potential_filename,
                        asi_filename,
                        elements_string=None,
                        add_to_job_name=None,
                        add_to_job_path=None):
        
        elements_string = elements_string or ' '.join(str(el) for el in self.structure.composition)
        input_string = f"""
        units       metal
        boundary    p p p
        atom_style  atomic
        box tilt large        
        read_data    {self.data_filename}
    
        pair_style  pace/extrapolation
        pair_coeff  * * {potential_filename} {asi_filename} {elements_string}
        
        fix pace_gamma all pair 10 pace/extrapolation gamma 1   
        compute max_pace_gamma all reduce max f_pace_gamma      
                     
        variable dump_skip equal "c_max_pace_gamma < 5"         
        dump pace_dump all custom 20 extrapolative_structures.dump id type x y z f_pace_gamma
        dump_modify pace_dump skip v_dump_skip
        
        variable max_pace_gamma equal c_max_pace_gamma         
        fix extreme_extrapolation all halt 10 v_max_pace_gamma > 25
         
        thermo 500    
        timestep 0.001
        
        # test     
        velocity all create 100.0 42
        fix 1 all nvt temp 50 300 0.1 
        run 10000
        unfix 1
        
        fix 1 all npt temp 300 300 0.1 aniso 0 10000 1
        run 20000
        unfix 1
        
        fix 1 all npt temp 300 300 0.1 aniso 10000 0 1
        run 20000
        unfix 1
        
        fix 1 all nvt temp 300 2000 0.1 
        run 68000
        unfix 1
        
        fix 1 all nvt temp 2000 300 0.1 
        run 68000
        unfix 1
                                
        fix 1 all box/relax aniso 0.0
        minimize  0 1.0e-8 10000 1000000
        unfix 1
        
        write_data out.data
        """
        self.lammps_input = LammpsInput.from_string(input_string)
        return self.get_lammpsjob(add_to_job_name=add_to_job_name,add_to_job_path=add_to_job_path)
            



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May  8 14:44:42 2025

@author: villa
"""
import os
import os.path as op
import warnings

from pymatgen.io.lammps.data import LammpsData

from pynter import SETTINGS
from pynter.hpc.slurm import JobSettings
from pynter.jobs.lammps.lammps_jobs import LammpsJob
from pynter.lammps.inputs import LammpsInput

class DefaultInputs:
    
    def get_lammps_input_default(input_string=None):
        input_string = SETTINGS['lammps']['input_string_default']
        if input_string:
            return LammpsInput.from_string(input_string)
        else:
            warnings.warn('Default LAMMPS input string is empty')
            return
    
    def get_lammps_data(self,structure,atom_style='atomic'):
        return LammpsData.from_structure(structure=structure,atom_style=atom_style)
        


class DefaultJobSettings:
    
    def __init__(self,
                 lammps_exe=None,
                 lammps_sbatch=None,
                 modules=None,
                 lines_before_srun=None,
                 lines_after_srun=None):
        """
        Class to handle LAMMPS specific part of sbatch submission script.
        Args that are not provided are taken from lammps.yml file in ~/.pynter.

        Parameters
        ----------
        lammps_exe : (str)
            Path to lammps executable. 
        lammps_sbatch : (dict)
            Sbatch dictionary tailored for lammps.
        modules : (list)
            Name of modules to load with "module load".
        lines_before_srun : (list)
            Script lines to be placed before the "srun" command.
        lines_after_srun : (list)
            Script lines to be placed before the "srun" command.
        """
        
        default_settings = SETTINGS['lammps']['job_settings_default']
        self.lammps_exe = lammps_exe or default_settings['lammps_exe']
        self.lammps_sbatch = lammps_sbatch or default_settings['lammps_sbatch']
        self.modules = modules or default_settings['modules']
        self.lines_before_srun = lines_before_srun or default_settings['lines_before_srun']
        self.lines_after_srun = lines_after_srun or default_settings['lines_after_srun']
    
    @property
    def lammps_script_lines(self):
        if self.modules:
            lines = self.get_modules_lines()
        else:
            lines = []
        if self.lines_before_srun:
            lines += self.lines_before_srun
        if self.lammps_exe:
            lines += [self.get_srun_line()]
        if self.lines_after_srun:
            lines += self.lines_after_srun
        return lines
        
    def get_modules_lines(self):
        if self.modules:
            return ['module purge'] + [f'module load {module}' for module in self.modules] 

    def get_srun_line(self):
        return 'srun %s' %self.lammps_exe

    def get_updated_job_settings(self,job_settings):
        """
        Get a copy of a JobSettings object updated with lammps specific settings
        """
        job_settings = job_settings.copy()
        for k,v in self.lammps_sbatch.items():
            job_settings[k] = v
        lines = self.lammps_script_lines
        job_settings['script_lines'] += lines
        return job_settings
    

input_default_filename = SETTINGS['lammps']['input_filename']
data_default_filename = SETTINGS['lammps']['data_filename']
    

class InputSets:
    
    def __init__(self,path=None,
                 structure=None,
                 input_string=None,
                 atom_style='atomic',
                 input_filename=input_default_filename,
                 data_filename=data_default_filename,
                 job_settings=None,
                 name=None,
                 add_parent_folder=False):
        """
        Parameters
        ----------
        path : (str)
            Main path for the scheme. If None the current work dir is used. The default is None.

        job_settings : (Dict), optional
            Dictionary with job settings to create job script, parameters are defined in JobSettings class function. The default is None.\n
            If job_settings is None, the 'job-name' key will be added, the value is the 'name' argument if provided, if 'name' arg is \n
            None the value will be: 'no_name'.
        name : (str), optional
            Name for the system to set up scheme for. The default is None.
        add_parent_folder : (bool), optional
            Add folder to the path names like the name of the InputSets. Default is False.
        """
        self.path = op.abspath(path) if path else os.getcwd()
        self.atom_style = atom_style
        self.input_string = input_string
        self.input_string_original = input_string
        self.input_filename = input_filename
        self.data_filename = data_filename
        
        
        if structure:
            self.structure = structure   
            self.lammps_data = DefaultInputs().get_lammps_data(structure=structure,atom_style=atom_style)
        else:
            self.lammps_data = None
            warnings.warn('Structure is not provided explicitly, make sure is defined in input string')
            
        if job_settings:
            self.job_settings = JobSettings(**job_settings)
        elif name:
            self.job_settings = JobSettings(sbatch={'job-name':name})
        else:
            self.job_settings = JobSettings(sbatch={'job-name':'no_name'})
        self.name = name if name != None else self.job_settings['job-name'] 
        
        if self.name: # this return false if name is '', the previuos line considers only if name is None
            self.job_settings['job-name'] = self.name
        
        self.job_settings = DefaultJobSettings().get_updated_job_settings(self.job_settings)    

        if add_parent_folder:
            self.path = op.join(self.path,self.name)
            
    
    def get_lammpsjob(self,add_to_job_name=None,add_to_job_path=None):
        """
        Generate LammpsJob object from the input settings of the class.

        Parameters
        ----------
        add_to_job_name : (str), optional
            String to be added to 'name' key in job_settings dictionary and to name attribute of LammpsJob.
        add_to_job_path : (str), optional
            String to be added to self.path. The complete path will be input in 'path' arg of LammpsJob.

        Returns
        -------
        LammpsJob object
        """        
        job_settings = self.job_settings.copy()
        
        if add_to_job_name:
            job_settings['job-name'] = '_'.join([job_settings['job-name'],add_to_job_name])        
            jobname = '_'.join([self.name,add_to_job_name])
        else:
            jobname = self.name
            
        if add_to_job_path:
            jobpath = op.join(self.path,add_to_job_path)
        else:
            jobpath = self.path
        
        if self.input_string:
            self.lammps_input = LammpsInput.from_string(self.input_string) 
        else:
            self.lammps_input = DefaultInputs().get_lammps_input_default()
        inputs = {'inp':self.lammps_input}
        if self.lammps_data:
            inputs['data'] = self.lammps_data
            
        job_script_filename = job_settings['filename'] if 'filename' in job_settings.keys() else None
        lammpsjob = LammpsJob(path=jobpath,
                              inputs=inputs,
                              job_settings=job_settings,
                              job_script_filename=job_script_filename,
                              name=jobname,
                              input_filename=self.input_filename,
                              data_filename=self.data_filename)
                              
        return lammpsjob  
    
  
    def npt(self,add_to_job_name=None,add_to_job_path=None,**kwargs):
        if self.input_string_original:
            warnings.warn('You are overriding your input in "input_string"')
        self.input_string = InputPresets().npt(**kwargs)
        return self.get_lammpsjob(add_to_job_name=add_to_job_name,add_to_job_path=add_to_job_path)
    
    def nvt(self,add_to_job_name=None,add_to_job_path=None,**kwargs):
        if self.input_string_original:
            warnings.warn('You are overriding your input in "input_string"')
        self.input_string = InputPresets().nvt(**kwargs)
        return self.get_lammpsjob(add_to_job_name=add_to_job_name,add_to_job_path=add_to_job_path)
        


class InputPresets:
    
            
    def npt(self,
            pair_style='',
            pair_coeff='',
            T=300,
            P=1,
            dt=1,
            nstepout=100,
            nstepdump=100,
            nsteprun=10000):
        lines = [
            "units metal",
            "boundary p p p",
            "atom_style atomic",
            "read_data structure.data",
            "",
            "pair_style %s" % pair_style,
            "pair_coeff %s" % pair_coeff,
            "",
            "variable T equal %s" % T,
            "variable P equal %s" %(P*1.01325), # pressure from atm to bar
            "variable dt equal %s" %(dt*0.001), # timestep from fs to ps,
            "",
            "timestep ${dt}",
            "variable T_init equal 2*$T",
            "velocity all create ${T_init} 4928459 rot yes dist gaussian",
            "",
            "fix             NPT all npt temp $T $T $(100.0*dt) iso $P $P $(1000.0*dt)",
            "",
            "variable nstepout equal %s" % nstepout,
            "variable nstepdump equal %s" % nstepdump,
            "variable nsteprun equal %s" % nsteprun,
            "",
            "thermo          ${nstepout}",
            "thermo_style    custom step temp pe ke etotal",
            "dump            myDump all custom ${nstepdump} structure.dump id type x y z",
            "run             ${nsteprun}",
        ]

    
        return '\n'.join(lines)
    
    
    def nvt(self,
            pair_style='',
            pair_coeff='',
            T=300,
            dt=1,
            nstepout=100,
            nstepdump=100,
            nsteprun=10000):
        lines = [
            "units metal",
            "boundary p p p",
            "atom_style atomic",
            "read_data structure.data",
            "",
            "pair_style %s" % pair_style,
            "pair_coeff %s" % pair_coeff,
            "",
            "variable T equal %s" % T,
            "variable dt equal %s" %(dt*0.001), # timestep from fs to ps
            "",
            "timestep ${dt}",
            "variable T_init equal 2*$T",
            "velocity all create ${T_init} 4928459 rot yes dist gaussian",
            "",
            "fix             NVT all nvt temp $T $T $(100.0*dt)",
            "",
            "variable nstepout equal %s" % nstepout,
            "variable nstepdump equal %s" % nstepdump,
            "variable nsteprun equal %s" % nsteprun,
            "",
            "thermo          ${nstepout}",
            "thermo_style    custom step temp pe ke etotal",
            "dump            myDump all custom ${nstepdump} structure.dump id type x y z",
            "run             ${nsteprun}",
        ]
    
        return '\n'.join(lines)